// effects-weed.c
// LiVES (lives-exe)
// (c) G. Finch 2005
// Released under the GPL 2.0
// see file ../COPYING for licensing details


#include <dlfcn.h>

#include "main.h"
#include "weed.h"
#include "weed-utils.h"
#include "effects-weed.h"

///////////////////////////////////

#include "callbacks.h"
#include "support.h"
#include "rte_window.h"
#include "resample.h"

///////////////////////////////////////////////////////////////////////
void *w_memset (void *s, int c, size_t n) {return memset(s,c,n);}
void *w_memcpy  (void *dest, const void *src, size_t n) {return memcpy(dest,src,n);}

weed_malloc_f weed_malloc=(weed_malloc_f)&g_malloc;
weed_free_f weed_free=(weed_free_f)&g_free;
weed_memset_f weed_memset=&w_memset;
weed_memcpy_f weed_memcpy=&w_memcpy;

////////////////////////////////////////////////////////////////////////

int weed_leaf_host_set_func (weed_plant_t *plant, const char *key, int seed_type, int num_elems, void *value) {
  // this is the host version, we will pass a version to plugins with caller set to WEED_CALLER_PLUGIN
  //printf("host set leaf %s\n",key);
  return weed_leaf_set_caller (plant,key,seed_type,num_elems,value,WEED_CALLER_HOST);
}

int weed_leaf_plugin_set_func (weed_plant_t *plant, const char *key, int seed_type, int num_elems, void *value) {
  // this is the plugin version, we will pass a version to plugins with caller set to WEED_CALLER_PLUGIN
  //printf("plugin set leaf %s\n",key);
  return weed_leaf_set_caller (plant,key,seed_type,num_elems,value,WEED_CALLER_PLUGIN);
}

weed_leaf_set_f weed_leaf_set=&weed_leaf_host_set_func; // host version
weed_leaf_set_f weed_leaf_plugin_set=&weed_leaf_plugin_set_func; // plugin version


////////////////////////////////////////////////////////////////////////////

static void weed_add_plant_flags (weed_plant_t *plant, int flags) {
  char **leaves=weed_plant_list_leaves(plant);
  int i,currflags;

  for (i=0;leaves[i]!=NULL;i++) {
    currflags=flags;
    if (flags&WEED_LEAF_READONLY_PLUGIN&&!strncmp(leaves[i],"plugin_",7)) currflags^=WEED_LEAF_READONLY_PLUGIN;
    weed_leaf_set_flags(plant,leaves[i],weed_leaf_get_flags(plant,leaves[i])|currflags);
    weed_free(leaves[i]);
  }
  weed_free(leaves);
}


static void weed_clear_plant_flags (weed_plant_t *plant, int flags) {
  char **leaves=weed_plant_list_leaves(plant);
  int i;

  for (i=0;leaves[i]!=NULL;i++) {
    weed_leaf_set_flags(plant,leaves[i],(weed_leaf_get_flags(plant,leaves[i])|flags)^flags);
    weed_free(leaves[i]);
  }
  weed_free(leaves);
}



weed_plant_t *weed_bootstrap_func (void **value, int num_versions, int *plugin_versions) {
  int host_api_version=100;
  weed_plant_t *host_info=weed_plant_new(WEED_PLANT_HOST_INFO);

  weed_leaf_get_f wlg=&weed_leaf_get;
  weed_plant_new_f wpn=&weed_plant_new;
  weed_plant_list_leaves_f wpll=&weed_plant_list_leaves;
  weed_leaf_num_elements_f wlne=&weed_leaf_num_elements;
  weed_leaf_element_size_f wles=&weed_leaf_element_size;
  weed_leaf_seed_type_f wlst=&weed_leaf_seed_type;
  weed_leaf_get_flags_f wlgf=&weed_leaf_get_flags;

  if (num_versions!=1) return NULL;

  if (plugin_versions[0]!=host_api_version) return NULL;

  // plugin version is 100, good.
  weed_memcpy(value,&wlg,sizeof(void *)); // bootstrap weed_leaf_get

  weed_set_int_value(host_info,"agreed_version",host_api_version);
  weed_set_voidptr_value(host_info,"weed_leaf_get_func",wlg);
  weed_set_voidptr_value(host_info,"weed_leaf_set_func",weed_leaf_plugin_set);
  weed_set_voidptr_value(host_info,"weed_plant_new_func",wpn);
  weed_set_voidptr_value(host_info,"weed_plant_list_leaves_func",wpll);
  weed_set_voidptr_value(host_info,"weed_leaf_num_elements_func",wlne);
  weed_set_voidptr_value(host_info,"weed_leaf_element_size_func",wles);
  weed_set_voidptr_value(host_info,"weed_leaf_seed_type_func",wlst);
  weed_set_voidptr_value(host_info,"weed_leaf_get_flags_func",wlgf);
  weed_set_voidptr_value(host_info,"weed_malloc_func",weed_malloc);
  weed_set_voidptr_value(host_info,"weed_free_func",weed_free);
  weed_set_voidptr_value(host_info,"weed_memset_func",weed_memset);
  weed_set_voidptr_value(host_info,"weed_memcpy_func",weed_memcpy);

  weed_add_plant_flags(host_info,WEED_LEAF_READONLY_PLUGIN);

  return host_info;
}

//////////////////////////////////////////////////////////////////////////////
// filter library functions


int weed_filter_categorise (int in_channels, int out_channels) {
  // TODO - add is_convertor, optional non-alpha in/out channels
  if (in_channels==0&&out_channels>0) return 1; // generator
  if (out_channels>1) return 7; // splitter : use optional non-alpha out_channels
  if (in_channels>2&&out_channels==1) return 5; // compositor : use optional non-alpha in_channels
  if (in_channels==2&&out_channels==1) return 2; // transition
  if (in_channels==1&&out_channels==1) return 3; // filter
  if (in_channels>0&&out_channels==0) return 6; // tap : use optional non-alpha in_channels
  if (in_channels==0&&out_channels==0) return 4; // utility
  return 0;
}


////////////////////////////////////////////////////////////////////////

#define MAX_WEED_FILTERS FX_KEYS_MAX
#define MAX_WEED_INSTANCES FX_KEYS_MAX

#define MAX_MODES_PER_KEY 8

// fixed size for generators - TODO - make configurable
#define DEF_GEN_WIDTH 640
#define DEF_GEN_HEIGHT 480

// store keys so we now eg, which rte mask entry to xor when deiniting
static int fg_generator_key;
static int bg_generator_key;

// store modes too
static int fg_generator_mode;
static int bg_generator_mode;

// generators to start on playback, because of a problem in libvisual 
// - we must start generators after starting audio
static int bg_gen_to_start;
static int fg_gen_to_start;

// store the clip, this can sometimes get lost
static gint fg_generator_clip;


//////////////////////////////////////////////////////////////////////

static weed_plant_t *weed_filters[MAX_WEED_FILTERS]; // array of filter_classes
static weed_plant_t *weed_instances[MAX_WEED_INSTANCES]; // array of filter_instances


// each 'hotkey' controls n instances, selectable as 'modes' or banks
static int key_to_instance[FX_KEYS_MAX>0?FX_KEYS_MAX:1][MAX_MODES_PER_KEY];
static int key_to_fx[FX_KEYS_MAX>0?FX_KEYS_MAX:1][MAX_MODES_PER_KEY];
static int key_modes[FX_KEYS_MAX>0?FX_KEYS_MAX:1];

static int next_free_instance=0; // we use an arena of instances

// count of how many filters we have loaded
static gint num_weed_filters;

gchar *hashnames[MAX_WEED_FILTERS];


/////////////////// LiVES event system /////////////////

static event_t *init_events[FX_KEYS_MAX_VIRTUAL];
static event_t * effect_map[FX_KEYS_MAX>0?FX_KEYS_MAX+1:1];
static int next_free_key;
static int key; // for assigning loaded filters

////////////////////////////////////////////////////////////////////

static inline gint step_val(gint val, gint step) {
  gint ret=(gint)(val/step+.5)*step;
   return ret==0?step:ret;
}


static void create_effect_map (void) {
  // here we create an effect map which defines the order in which effects are applied to a frame stack
  // this is done during recording, the keymap is from mainw->rte which is a bitmap of effect keys
  // keys are applied here from smallest (ctrl-1) to largest (virtual key ctrl-64)

  // this is transformed into effect_map which holds init_events 
  // these pointers are then stored in an effect_map event

  // what we actually point to are the init_events for the effects. The init_events are stored when we 
  // init an effect

  // during rendering we read the effect_map event, and retrieve the new key, which is at that time 
  // held in the 
  // "host_tag" property of the init_event, and we apply our effects
  // (which are then bound to virtual keys >64)

  // [note] that we can do cool things, like mapping the same instance multiple times (though it will always
  // apply itself to the same in/out tracks

  // we don't need to worry about free()ing init_events, since they will be free'd 
  // when the instance is deinited

  int count=0,i;

  for (i=0;i<64;i++) if (mainw->rte&(GU641<<i)&&enabled_in_channels (key_to_instance[i][key_modes[i]],FALSE)>0) effect_map[count++]=init_events[i];
  effect_map[count]=NULL; // marks the end of the effect map
}


void add_effect_deinit_events (void) {
  // during rendering we use the "keys" 64 -> FX_KEYS_MAX
  // here we add effect_deinit events to an event_list
  int i;

  for (i=0;i<64;i++) {
    if (init_events[i]!=NULL) {
      add_effect_deinit_event (mainw->event_list,mainw->currticks-mainw->origticks,init_events[i]);
      init_events[i]=NULL;
    }
  }
  // add an empty effect_map event (in case more frames are added)
  create_effect_map(); // we create effect_map event_t * array with ordered effects
  add_effect_map_event (mainw->event_list,mainw->currticks-mainw->origticks,effect_map);
}


void add_effect_init_events (void) {
  // during rendering we use the "keys" 64 -> FX_KEYS_MAX
  // here we are about to start playback, and we add init events for every effect which is switched on
  // we add the init events with a timecode of 0
  int i;
  gint idx;
  event_list_t *event_list;

  for (i=0;i<64;i++) {
    if ((idx=key_to_instance[i][key_modes[i]])!=-1&&enabled_in_channels (idx,FALSE)>0) {
      event_list=append_effect_init_event (mainw->event_list,0,key_to_fx[i][key_modes[i]]);
      if (mainw->event_list==NULL) mainw->event_list=event_list;
      init_events[i]=mainw->event_list->last;
    }
  }
  // add an empty effect_map event (in case more frames are added)
  create_effect_map(); // we create effect_map event_t * array with ordered effects
  if (effect_map[0]!=NULL) add_effect_map_event (mainw->event_list,0,effect_map);
}


// check if palette is in the palette_list for the channel template
static gboolean check_weed_palette_list (weed_plant_t *chantmpl, int palette) {
  int i,error,num_palettes;
  int *palettes;

  if (chantmpl==NULL) return FALSE;
  if ((num_palettes=weed_leaf_num_elements(chantmpl,"palette_list"))==0) return FALSE;

  palettes=weed_get_int_array(chantmpl,"palette_list",&error);

  for (i=0;i<num_palettes;i++) {
    if (palettes[i]==palette) {
      weed_free(palettes);
      return TRUE;
    }
  }
  weed_free(palettes);
  return FALSE;
}

static void set_channel_size (weed_plant_t *channel, gint width, gint height, int numplanes, int *rowstrides) {
  int error;
  int max;
  weed_plant_t *chantmpl=weed_get_plantptr_value(channel,"template",&error);
  // note: rowstrides is just a guess, we will set the actual value when we come to process the effect


  if (weed_plant_has_leaf(chantmpl,"width")) width=weed_get_int_value(chantmpl,"width",&error);
  else if (weed_plant_has_leaf(chantmpl,"hstep")) width=step_val(width,weed_get_int_value(chantmpl,"hstep",&error));
  if (weed_plant_has_leaf(chantmpl,"maxwidth")) {
    max=weed_get_int_value(chantmpl,"maxwidth",&error);
    if (width>max) width=max;
  }
  weed_set_int_value(channel,"width",width);

  if (weed_plant_has_leaf(chantmpl,"height")) height=weed_get_int_value(chantmpl,"height",&error);
  else if (weed_plant_has_leaf(chantmpl,"vstep")) height=step_val(height,weed_get_int_value(chantmpl,"vstep",&error));
  if (weed_plant_has_leaf(chantmpl,"maxheight")) {
    max=weed_get_int_value(chantmpl,"maxheight",&error);
    if (height>max) height=max;
  }
  weed_set_int_value(channel,"height",height);

  if (rowstrides!=NULL) weed_set_int_array(channel,"rowstrides",numplanes,rowstrides);
  
}


static gboolean rowstrides_differ(int n1, int *n1_array, int n2, int *n2_array) {
  // returns TRUE if the rowstrides differ
  int i;

  if (n1!=n2) return TRUE;
  for (i=0;i<n1;i++) if (n1_array[i]!=n2_array[i]) return TRUE;
  return FALSE;
}


static gboolean align (void **pixel_data, size_t alignment, int numplanes, int height, int *rowstrides) {
  // returns TRUE on success
  int i;
  for (i=0;i<numplanes;i++) {
    if (((gulong)(pixel_data[i]))%alignment==0) continue;
#ifdef HAVE_POSIX_MEMALIGN
    else {
      int memerror;
      void *new_pixel_data;
      if ((memerror=posix_memalign(&new_pixel_data,alignment,height*rowstrides[i]))) return FALSE;
      memcpy(new_pixel_data,pixel_data[i],height*rowstrides[i]);
      g_free(pixel_data[i]);
      pixel_data[i]=new_pixel_data;
    }
#else
    return FALSE;
#endif
  }
  return TRUE;
}


static weed_plant_t **backup_param_values (weed_plant_t *inst) {
  int num_params=weed_leaf_num_elements(inst,"in_parameters");
  weed_plant_t **backup=NULL;

  if (num_params>0) {
    int error,i;
    weed_plant_t **in_params=weed_get_plantptr_array(inst,"in_parameters",&error);
    backup=weed_malloc(num_params*sizeof(weed_plant_t *));
    for (i=0;i<num_params;i++) {
      backup[i]=weed_plant_new(WEED_PLANT_PARAMETER);
      weed_leaf_copy(in_params[i],"value",backup[i],"value");
    }
  }
  return backup;
}

static void restore_param_values (weed_plant_t *inst, weed_plant_t **backup) {
  // this function will free elements in backup, and backup itself
  int num_params=weed_leaf_num_elements(inst,"in_parameters");

  if (num_params>0) {
    int error,i;
    weed_plant_t **in_params=weed_get_plantptr_array(inst,"in_parameters",&error);
    for (i=0;i<num_params;i++) {
      weed_leaf_copy(backup[i],"value",in_params[i],"value");
      weed_plant_free(backup[i]);
    }
    weed_free(backup);
  }
}


static void set_param_gui_readonly (weed_plant_t *inst) {
  int num_params,error,i;
  weed_plant_t **params,*gui,*ptmpl;

  num_params=weed_leaf_num_elements(inst,"in_parameters");
  if (num_params>0) {
    params=weed_get_plantptr_array(inst,"in_parameters",&error);
    for (i=0;i<num_params;i++) {
      ptmpl=weed_get_plantptr_value(params[i],"template",&error);
      if (weed_plant_has_leaf(ptmpl,"gui")) {
	gui=weed_get_plantptr_value(ptmpl,"gui",&error);
	weed_add_plant_flags(gui,WEED_LEAF_READONLY_PLUGIN);
      }
    }
    weed_free(params);
  }

  num_params=weed_leaf_num_elements(inst,"out_parameters");
  if (num_params>0) {
    params=weed_get_plantptr_array(inst,"out_parameters",&error);
    for (i=0;i<num_params;i++) {
      ptmpl=weed_get_plantptr_value(params[i],"template",&error);
      if (weed_plant_has_leaf(ptmpl,"gui")) {
	gui=weed_get_plantptr_value(ptmpl,"gui",&error);
	weed_add_plant_flags(gui,WEED_LEAF_READONLY_PLUGIN);
      }
    }
    weed_free(params);
  }
}

static void set_param_gui_readwrite (weed_plant_t *inst) {
  int num_params,error,i;
  weed_plant_t **params,*gui,*ptmpl;

  num_params=weed_leaf_num_elements(inst,"in_parameters");
  if (num_params>0) {
    params=weed_get_plantptr_array(inst,"in_parameters",&error);
    for (i=0;i<num_params;i++) {
      ptmpl=weed_get_plantptr_value(params[i],"template",&error);
      if (weed_plant_has_leaf(ptmpl,"gui")) {
	gui=weed_get_plantptr_value(ptmpl,"gui",&error);
	weed_clear_plant_flags(gui,WEED_LEAF_READONLY_PLUGIN);
      }
    }
    weed_free(params);
  }

  num_params=weed_leaf_num_elements(inst,"out_parameters");
  if (num_params>0) {
    params=weed_get_plantptr_array(inst,"out_parameters",&error);
    for (i=0;i<num_params;i++) {
      ptmpl=weed_get_plantptr_value(params[i],"template",&error);
      if (weed_plant_has_leaf(ptmpl,"gui")) {
	gui=weed_get_plantptr_value(ptmpl,"gui",&error);
	weed_clear_plant_flags(gui,WEED_LEAF_READONLY_PLUGIN);
      }
    }
    weed_free(params);
  }
}







gint weed_apply_instance (weed_plant_t *inst, weed_plant_t *init_event, weed_plant_t **layers, int opwidth, int opheight, weed_timecode_t tc) {
  // here we:
  // get our in_tracks and out_tracks that map inst channels to layers

  // clear "disabled" if we have non-zero frame and there is no "disabled" in template
  // if we have a zero frame, set "disabled" if "optional", otherwise we cannot apply the filter

  // set channel timecodes

  // pull pixel_data (unless it is there already)

  // try to set channel width,height,current_palette, to match largest of in layers, or opwidth, opheight, whichever is smallest, reiniting if necessary


  // if palette is wrong, convert palette in the layer

  // if width and height are wrong, resize in the layer

  // apply the effect, put result in output layer, set layer palette, width, height, rowstrides

  // if filter does not support inplace, we must create a new pixel_data; this will then replace the original layer


  // TODO ** - handle return errors


  int num_in_tracks,num_out_tracks;
  int *in_tracks,*out_tracks;
  int error,i,j;
  weed_plant_t *filter=weed_get_plantptr_value(inst,"filter_class",&error);
  weed_plant_t *layer;
  int maxinwidth=0;
  int maxinheight=0;
  weed_plant_t **in_channels,**out_channels,*channel,*chantmpl;
  int frame;
  int inwidth,inheight,inpalette,outpalette,channel_flags,palette;
  int outwidth,outheight;
  gboolean needs_reinit=FALSE,inplace=FALSE;
  int incwidth,incheight,numplanes=0,width,height;
  int *rowstrides;
  void **pixel_data;
  weed_process_f process_func;
  weed_plant_t *def_channel=NULL;
  gboolean rowstrides_changed;
  int nchr;
  int *layer_rows=NULL,*channel_rows;

  // here, in_tracks and out_tracks map our layers to in_channels and out_channels in the filter

  if (!weed_plant_has_leaf(inst,"in_channels")||(in_channels=weed_get_plantptr_array(inst,"in_channels",&error))==NULL) return FILTER_ERROR_NO_IN_CHANNELS;

  if (init_event==NULL) {
    num_in_tracks=weed_leaf_num_elements(inst,"in_channels"); // TODO - **
    if (num_in_tracks==0) return FILTER_ERROR_NO_IN_CHANNELS;
    if (num_in_tracks==2&&layers[1]==NULL) return FILTER_ERROR_MISSING_LAYER; // TODO ** !!
    in_tracks=weed_malloc(2*sizint);
    in_tracks[0]=0;
    in_tracks[1]=1;
    num_out_tracks=1;
    out_tracks=weed_malloc(sizint);
    out_tracks[0]=0;
  }
  else {
    num_in_tracks=weed_leaf_num_elements(init_event,"in_tracks");
    in_tracks=weed_get_int_array(init_event,"in_tracks",&error);
    num_out_tracks=weed_leaf_num_elements(init_event,"out_tracks");
    out_tracks=weed_get_int_array(init_event,"out_tracks",&error);
  }

  out_channels=weed_get_plantptr_array(inst,"out_channels",&error);

  // TODO - handle case where in_tracks[i] > than num layers
  // either we temporarily disable the channel, or we can't apply the filter

  for (i=0;i<num_in_tracks;i++) {
    layer=layers[in_tracks[i]];
    if (layer==NULL) {
      weed_free(in_tracks);
      weed_free(out_tracks);
      weed_free(in_channels);
      weed_free(out_channels);
      return FILTER_ERROR_MISSING_LAYER;
    }
    //channel=get_enabled_channel(in_channels,i);
    if (weed_get_voidptr_value(layer,"pixel_data",&error)==NULL) {
      frame=weed_get_int_value(layer,"frame",&error);
      if (frame==0) {
	
	// TODO - temp disable channels if we can
	
	//chantmpl=weed_plant_get_leaf(channel,"template",&error);
	// blank frame, see if we can disable this channel temporarily
	//if (weed_plant_has_leaf(chantmpl,"optional")) {
	//if (!weed_plant_has_leaf(channel,"disabled")) needs_reinit=TRUE;
	//weed_set_boolean_value(channel,"disabled",WEED_TRUE);
	// we will renable this later by setting "disabled" to WEED_FALSE
	//}
	//else return FALSE; // missing frame, so we cannot apply the filter
	//}
	weed_free(in_tracks);
	weed_free(out_tracks);
	weed_free(in_channels);
	weed_free(out_channels);
	return FILTER_ERROR_BLANK_FRAME;
      }
    }
  }

  if (get_enabled_channel(in_channels,0)==NULL) {
    // we process generators elsewhere
    weed_free(in_tracks);
    weed_free(out_tracks);
    weed_free(in_channels);
    weed_free(out_channels);
    return FILTER_ERROR_NO_IN_CHANNELS;
  }

  // pull frames for tracks

  for (i=0;i<num_in_tracks;i++) {
    layer=layers[in_tracks[i]];
    if (weed_get_voidptr_value(layer,"pixel_data",&error)==NULL) {
      // pull_frame will set pixel_data,width,height,current_palette and rowstrides
      if (!pull_frame(layer,prefs->image_ext,tc)) return FILTER_ERROR_MISSING_FRAME;
    }
    if ((inwidth=weed_get_int_value(layer,"width",&error))>maxinwidth) maxinwidth=inwidth;
    if ((inheight=weed_get_int_value(layer,"height",&error))>maxinheight) maxinheight=inheight;
  }

  if (maxinwidth<opwidth) opwidth=maxinwidth;
  if (maxinheight<opheight) opheight=maxinheight;

  // set up in_channels from layers
  for (i=0;i<num_in_tracks;i++) {
    // first we convert palettes
    layer=layers[in_tracks[i]];
    channel=get_enabled_channel(in_channels,i);
    if (def_channel==NULL) def_channel=channel;
    chantmpl=weed_get_plantptr_value(channel,"template",&error);
    palette=weed_get_int_value(layer,"current_palette",&error);
    inpalette=weed_get_int_value(channel,"current_palette",&error);
    channel_flags=weed_get_int_value(chantmpl,"flags",&error);

    // try to match palettes with first enabled in channel
    if (i>0&&!(channel_flags&WEED_CHANNEL_PALETTE_CAN_VARY)) palette=weed_get_int_value(def_channel,"current_palette",&error);

    if (palette!=inpalette) {
      // palette change needed; first try to change channel palette
      if (check_weed_palette_list(chantmpl,palette)) {
	weed_set_int_value(channel,"current_palette",palette);
	if (channel_flags&WEED_CHANNEL_REINIT_ON_PALETTE_CHANGE) needs_reinit=TRUE;
	inpalette=palette;
      }
    }

    if (weed_get_int_value(layer,"current_palette",&error)!=inpalette) {
      // otherwise convert layer palette
      if (!convert_layer_palette(layer,inpalette)) {
	weed_free(in_tracks);
	weed_free(out_tracks);
	weed_free(in_channels);
	weed_free(out_channels);
	return FILTER_ERROR_INVALID_PALETTE_CONVERSION;
      }
    }
  }

  // now we resize if necessary
  for (i=0;i<num_in_tracks;i++) {
    layer=layers[in_tracks[i]];
    channel=get_enabled_channel(in_channels,i);
    chantmpl=weed_get_plantptr_value(channel,"template",&error);

    incwidth=weed_get_int_value(channel,"width",&error);
    incheight=weed_get_int_value(channel,"height",&error);

    nchr=weed_leaf_num_elements(channel,"rowstrides");
    channel_rows=weed_get_int_array(channel,"rowstrides",&error);

    palette=weed_get_int_value(channel,"current_palette",&error);

    width=opwidth;
    height=opheight;

    channel_flags=weed_get_int_value(chantmpl,"flags",&error);

    inwidth=weed_get_int_value(layer,"width",&error);
    inheight=weed_get_int_value(layer,"height",&error);

    if (channel_flags&WEED_CHANNEL_SIZE_CAN_VARY) {
      width=inwidth;
      height=inheight;
    }

    // try to set channel to match layer, we might not be able to
    numplanes=weed_leaf_num_elements(layer,"rowstrides");
    rowstrides=weed_get_int_array(layer,"rowstrides",&error);
    
    set_channel_size(channel,width,height,numplanes,rowstrides);
    weed_free(rowstrides);

    height=weed_get_int_value(channel,"height",&error);
    width=weed_get_int_value(channel,"width",&error);

    if ((inwidth!=width)||(inheight!=height)) {
      // layer needs resizing
      if (prefs->pb_quality==PB_QUALITY_HIGH) {
	resize_layer(layer,width,height,GDK_INTERP_HYPER);
      }
      if (prefs->pb_quality==PB_QUALITY_MED) {
	resize_layer(layer,width,height,GDK_INTERP_BILINEAR);
      }
      else {
	resize_layer(layer,width,height,GDK_INTERP_NEAREST);
      }

      inwidth=weed_get_int_value(layer,"width",&error);
      inheight=weed_get_int_value(layer,"height",&error);
      if ((inwidth!=width)||(inheight!=height)) {
	weed_free(in_tracks);
	weed_free(out_tracks);
	weed_free(in_channels);
	weed_free(out_channels);
	return FILTER_ERROR_UNABLE_TO_RESIZE;
      }
    }

    numplanes=weed_leaf_num_elements(layer,"rowstrides");
    rowstrides=weed_get_int_array(layer,"rowstrides",&error);

    // after all resizing is done, we set the real rowstrides
    weed_set_int_array(channel,"rowstrides",numplanes,rowstrides);

    // check layer rowstrides against previous settings
    rowstrides_changed=rowstrides_differ(numplanes,rowstrides,nchr,channel_rows);
    weed_free(channel_rows);

    if ((rowstrides_changed||(incwidth!=width)||(incheight!=height))&&(channel_flags&WEED_CHANNEL_REINIT_ON_SIZE_CHANGE)) needs_reinit=TRUE;

    weed_set_int64_value(channel,"timecode",tc);
    pixel_data=weed_get_voidptr_array(layer,"pixel_data",&error);
 
    // align memory if necessary
    if (weed_plant_has_leaf(chantmpl,"alignment")) {
      int alignment=weed_get_int_value(chantmpl,"alignment",&error);
      align(pixel_data,alignment,numplanes,height,rowstrides);
      weed_set_voidptr_array(layer,"pixel_data",numplanes,pixel_data);
    }

    weed_free(rowstrides);
    weed_set_voidptr_array(channel,"pixel_data",numplanes,pixel_data);
    weed_free(pixel_data);
  }

  // set up our out channels
  for (i=0;i<num_out_tracks;i++) {
    channel=get_enabled_channel(out_channels,i);

    outwidth=weed_get_int_value(channel,"width",&error);
    outheight=weed_get_int_value(channel,"height",&error);

    weed_set_int64_value(channel,"timecode",tc);
    outpalette=weed_get_int_value(channel,"current_palette",&error);
    chantmpl=weed_get_plantptr_value(channel,"template",&error);
    channel_flags=weed_get_int_value(chantmpl,"flags",&error);
    nchr=weed_leaf_num_elements(channel,"rowstrides");
    channel_rows=weed_get_int_array(channel,"rowstrides",&error);

    if (def_channel!=NULL&&i==0&&(in_tracks[0]==out_tracks[0])) {
      if (channel_flags&WEED_CHANNEL_CAN_DO_INPLACE) {
	// ah, good, inplace
	palette=weed_get_int_value(def_channel,"current_palette",&error);
	if (check_weed_palette_list(chantmpl,palette)) {
	  weed_set_int_value(channel,"current_palette",palette);
	  if (outpalette!=palette&&(channel_flags&WEED_CHANNEL_REINIT_ON_PALETTE_CHANGE)) needs_reinit=TRUE;
	  width=weed_get_int_value(def_channel,"width",&error);
	  height=weed_get_int_value(def_channel,"height",&error);
	  weed_set_int_value(channel,"width",width);
	  weed_set_int_value(channel,"height",height);
	  weed_set_int_value(channel,"current_palette",palette);
	  numplanes=weed_leaf_num_elements(def_channel,"rowstrides");
	  layer_rows=weed_get_int_array(def_channel,"rowstrides",&error);
	  weed_set_int_array(channel,"rowstrides",numplanes,layer_rows);
	  pixel_data=weed_get_voidptr_array(def_channel,"pixel_data",&error);
	  weed_set_voidptr_array(channel,"pixel_data",numplanes,pixel_data);
	  weed_free(pixel_data);
	  weed_set_boolean_value(channel,"inplace",WEED_TRUE);
	  inplace=TRUE;
	}
      }
    }

    if (!inplace) {
      if (def_channel==NULL) def_channel=get_enabled_channel(out_channels,0);
      // try to match palettes with first enabled in channel
      palette=weed_get_int_value(def_channel,"current_palette",&error);
      
      if (palette!=outpalette) {
	// palette change needed; try to change channel palette
	if (check_weed_palette_list(chantmpl,palette)) {
	  weed_set_int_value(channel,"current_palette",palette);
	  if (channel_flags&WEED_CHANNEL_REINIT_ON_PALETTE_CHANGE) needs_reinit=TRUE;
	}
	else {
	  weed_free(in_tracks);
	  weed_free(out_tracks);
	  weed_free(in_channels);
	  weed_free(out_channels);
	  weed_free(channel_rows);
	  return FILTER_ERROR_INVALID_PALETTE_SETTINGS; // plugin author messed up...
	}
      }

      set_channel_size(channel,opwidth,opheight,1,NULL);
      create_empty_pixel_data(channel); // this will look at width, height, current_palette, and create an empty pixel_data and set rowstrides
      // and update width and height if necessary

      numplanes=weed_leaf_num_elements(channel,"rowstrides");
      layer_rows=weed_get_int_array(channel,"rowstrides",&error);
      // align memory if necessary
      if (weed_plant_has_leaf(chantmpl,"alignment")) {
	int alignment=weed_get_int_value(chantmpl,"alignment",&error);
	pixel_data=weed_get_voidptr_array(channel,"pixel_data",&error);
	height=weed_get_int_value(channel,"height",&error);
	align(pixel_data,alignment,numplanes,height,layer_rows);
	weed_set_voidptr_array(channel,"pixel_data",numplanes,pixel_data);
	weed_free(pixel_data);
      }
      weed_set_boolean_value(channel,"inplace",WEED_FALSE);
    }

    // check old rowstrides against current rowstrides
    rowstrides_changed=rowstrides_differ(nchr,channel_rows,numplanes,layer_rows);

    weed_free(channel_rows);
    weed_free(layer_rows);
      
    width=weed_get_int_value(channel,"width",&error);
    height=weed_get_int_value(channel,"height",&error);
    if ((rowstrides_changed||((outwidth!=width)||(outheight!=height)))&&(channel_flags&WEED_CHANNEL_REINIT_ON_SIZE_CHANGE)) needs_reinit=TRUE;
  }

  if (needs_reinit) {
    int num_params=weed_leaf_num_elements(inst,"in_parameters");
    // TODO - allow "gui" changes
    // TODO - backup and restore param values
    if (weed_plant_has_leaf(filter,"deinit_func")) {
      weed_deinit_f deinit_func=weed_get_voidptr_value(filter,"deinit_func",&error);
      if (deinit_func!=NULL) (*deinit_func)(inst);
    }
    if (weed_plant_has_leaf(filter,"init_func")) {
      weed_plant_t **backup_params=backup_param_values(inst);
      weed_init_f init_func=weed_get_voidptr_value(filter,"init_func",&error);
      if (init_func!=NULL) {
	set_param_gui_readwrite(inst);
	if ((*init_func)(inst)!=WEED_NO_ERROR) {
	weed_free(in_tracks);
	weed_free(out_tracks);
	weed_free(in_channels);
	weed_free(out_channels);
	for (j=0;j<num_params;j++) weed_plant_free(backup_params[j]);
	weed_free(backup_params);
	return FILTER_ERROR_COULD_NOT_REINIT;
	}
	set_param_gui_readonly(inst);
	restore_param_values(inst,backup_params);
      }
    }
    else needs_reinit=FALSE;
  }

  weed_set_double_value(inst,"fps",cfile->pb_fps);

  // TODO - redraw GUI

  //...finally we are ready to apply the filter
  process_func=weed_get_voidptr_value(filter,"process_func",&error);
  if ((*process_func)(inst,tc)==WEED_ERROR_PLUGIN_INVALID) {
    weed_free(in_tracks);
    weed_free(out_tracks);
    weed_free(in_channels);
    weed_free(out_channels);
    return FILTER_ERROR_MUST_RELOAD;
  }

  // TODO - handle process errors (WEED_ERROR_PLUGIN_INVALID)

  // now we write our out channels back to layers, leaving the palettes and sizes unchanged
  for (i=0;i<num_out_tracks;i++) {
    channel=get_enabled_channel(out_channels,i);
    if (weed_get_boolean_value(channel,"inplace",&error)==WEED_TRUE) continue;
    layer=layers[out_tracks[i]];
    numplanes=weed_leaf_num_elements(layer,"rowstrides");
    rowstrides=weed_get_int_array(layer,"rowstrides",&error);
    weed_set_int_array(layer,"rowstrides",numplanes,rowstrides);
    weed_free(rowstrides);
    numplanes=weed_leaf_num_elements(layer,"pixel_data");
    pixel_data=weed_get_voidptr_array(layer,"pixel_data",&error);
    for (j=0;j<numplanes;j++) g_free(pixel_data[j]);
    weed_free(pixel_data);
    numplanes=weed_leaf_num_elements(channel,"pixel_data");
    pixel_data=weed_get_voidptr_array(channel,"pixel_data",&error);
    weed_set_voidptr_array(layer,"pixel_data",numplanes,pixel_data);
    weed_free(pixel_data);
    weed_set_int_value(layer,"current_palette",weed_get_int_value(channel,"current_palette",&error));
    weed_set_int_value(layer,"width",weed_get_int_value(channel,"width",&error));
    weed_set_int_value(layer,"height",weed_get_int_value(channel,"height",&error));
  }
  // done...

  weed_free(in_tracks);
  weed_free(out_tracks);
  weed_free(in_channels);
  weed_free(out_channels);

  if (needs_reinit) return FILTER_ERROR_REINITED;

  return FILTER_NO_ERROR;
}





void weed_apply_effect_map (weed_plant_t **layers, weed_plant_t *filter_map, gint opwidth, gint opheight, weed_timecode_t tc) {
  int i,key,num_inst,error;
  weed_plant_t *instance;
  gchar *keystr;
  event_t **init_events;
  int filter_error;

  // this is called during rendering - we will have previously received a effect_map event and now we apply this to layers


  // layers will be a NULL terminated array of channels, each with two extra leaves: clip and frame
  // clip corresponds to a LiVES file in mainw->files. "pixel_data" will initially be NULL, we will pull this as necessary
  // and the effect output is written back to the layers

  // a frame number of 0 indicates a blank frame; if an effect gets one of these it will not process (except if the channel is optional, in which case the channel is disabled); the same is true for invalid track numbers - hence disabled channels which do not have disabled in the template are re-enabled when a frame is available

  // after all processing, we will display/output the first non-NULL layer. If all layers are NULL we generate a blank frame

  // finally, the output is resized to the display size

  // we can make some optimisations; we only resize and convert palettes as necessary

  // the channel sizes are set by the filter: all channels are all set to the display size, or to the size of the largest input layer (whichever is smallest). We attempt to do this, but some channels have fixed sizes. Then we resize "pixel_data" as necessary. 
  // If we fail to resize exactly, we go back and start again using the adjusted size. If we cannot resize to a fixed channel size, the filter cannot be applied.

  if (filter_map==NULL||!weed_plant_has_leaf(filter_map,"init_events")) return;
  if ((num_inst=weed_leaf_num_elements(filter_map,"init_events"))>0) {
    init_events=(event_t **)weed_get_voidptr_array(filter_map,"init_events",&error);
    for (i=0;i<num_inst;i++) {
      keystr=weed_get_string_value(init_events[i]->event,"host_tag",&error);
      key=atoi(keystr);
      weed_free(keystr);
      if (rte_key_valid (key)) {
	if (key_to_instance[key][key_modes[key]]==-1) continue;
	instance=weed_instances[key_to_instance[key][key_modes[key]]];
	if (instance==NULL) continue;
	filter_error=weed_apply_instance (instance,init_events[i]->event,layers,opwidth,opheight,tc);
	//if (filter_error!=FILTER_NO_ERROR) g_printerr("Render error was %d\n",filter_error);
      }
    }
    weed_free(init_events);
  }
}




GdkPixbuf *weed_apply_effects (weed_plant_t **layers, weed_plant_t *filter_map, gint opwidth, gint opheight, weed_timecode_t tc) {
  // TODO - update the param window with any out_param values which changed.
  // TODO - return a layer and not a GdkPixbuf

  int i,error;
  weed_plant_t *instance,*layer;
  int width,height;
  int filter_error;
  int output;

  if (mainw->is_rendering&&!(cfile->proc_ptr!=NULL&&mainw->preview)) {
    if (layers[0]!=NULL&&weed_get_int_value(layers[0],"clip",&error)!=mainw->scrap_file) {
      weed_apply_effect_map(layers, filter_map, opwidth, opheight, tc);
    }
  }

  // free playback: we will have here only one or two layers, and no effect_map.
  // Effects are applied in key order, in tracks are 0 and 1, out track is 0
  else {
    for (i=0;i<FX_KEYS_PHYSICAL;i++) {
      if (rte_key_valid(i)) {
	if (mainw->rte&(GU641<<i)) {
	  mainw->osc_block=TRUE;
	  if (key_to_instance[i][key_modes[i]]==-1) continue;
	  instance=weed_instances[key_to_instance[i][key_modes[i]]];
	  if (instance==NULL) continue;
	  filter_error=weed_apply_instance (instance,NULL,layers,opwidth,opheight,tc);
	  if (filter_error==FILTER_ERROR_REINITED) close_pwindow(i,key_modes[i],TRUE); // redraw our paramwindow
	  //if (filter_error!=FILTER_NO_ERROR) g_printerr("Render error was %d\n",filter_error);
	}
      }
    }
  }
  if (!mainw->is_rendering) mainw->osc_block=FALSE;

  // caller should free all layers, but here we will free all other pixel_data
  output=-1;

  for (i=0;layers[i]!=NULL;i++) {
    if (weed_get_voidptr_value(layers[i],"pixel_data",&error)!=NULL||weed_get_int_value(layers[i],"frame",&error)!=0) {
      if (output==-1) output=i;
      else {
	int j;
	void **pixel_data=weed_get_voidptr_array(layers[i],"pixel_data",&error);
	int numplanes=weed_leaf_num_elements(layers[i],"pixel_data");
	for (j=0;j<numplanes;j++) g_free(pixel_data[j]);
	weed_free(pixel_data);
      }
    }
  }

  if (output==-1) return gdk_pixbuf_new_blank(opwidth,opheight,WEED_PALETTE_BGR24);
  layer=layers[output];

  if (weed_get_voidptr_value(layer,"pixel_data",&error)==NULL) if (!pull_frame(layer,prefs->image_ext,tc)) return gdk_pixbuf_new_blank(opwidth,opheight,WEED_PALETTE_BGR24);

  convert_layer_palette(layer,WEED_PALETTE_BGR24);
  width=weed_get_int_value(layer,"width",&error);
  height=weed_get_int_value(layer,"height",&error);

  if (width!=opwidth||height!=opheight) {
    if (prefs->pb_quality==PB_QUALITY_HIGH) {
      resize_layer(layer,opwidth,opheight,GDK_INTERP_HYPER);
    }
    if (prefs->pb_quality==PB_QUALITY_MED) {
      resize_layer(layer,opwidth,opheight,GDK_INTERP_BILINEAR);
    }
    else {
      resize_layer(layer,opwidth,opheight,GDK_INTERP_NEAREST);
    }
  }
  return layer_to_pixbuf(layer);
}







/////////////////////////////////////////////////////////////////////////


static int check_weed_plugin_info (weed_plant_t *plugin_info) {
  // verify the plugin_info returned from the plugin
  // TODO - print descriptive errors
  weed_plant_t *host_info;
  int error;

  if (!weed_plant_has_leaf(plugin_info,"host_info")) return -1;
  if (!weed_plant_has_leaf(plugin_info,"version")) return -2;
  if (!weed_plant_has_leaf(plugin_info,"filters")) return -3;
  host_info=weed_get_plantptr_value(plugin_info,"host_info",&error); // TODO - check this

  return weed_leaf_num_elements(plugin_info,"filters");
}


gint enabled_in_channels (gint idx, gboolean use_template) {
  weed_plant_t *filter,**channels=NULL;
  gint enabled=0;
  int num_channels,i,error;

  if (use_template) {
    filter=weed_filters[idx];
    num_channels=weed_leaf_num_elements(filter,"in_channel_templates");
    if (num_channels>0) channels=weed_get_plantptr_array(filter,"in_channel_templates",&error);
  }
  else {
    filter=weed_instances[idx];
    num_channels=weed_leaf_num_elements(filter,"in_channels");
    if (num_channels>0) channels=weed_get_plantptr_array(filter,"in_channels",&error);
  }

  for (i=0;i<num_channels;i++) if (!weed_plant_has_leaf(channels[i],"disabled")||weed_get_boolean_value(channels[i],"disabled",&error)!=WEED_TRUE) enabled++;

  if (channels!=NULL) weed_free(channels);

  return enabled;
}


gint enabled_out_channels (gint idx, gboolean use_template) {
  weed_plant_t *filter,**channels=NULL;
  gint enabled=0;
  int num_channels,i,error;

  if (use_template) {
    filter=weed_filters[idx];
    num_channels=weed_leaf_num_elements(filter,"out_channel_templates");
    if (num_channels>0) channels=weed_get_plantptr_array(filter,"out_channel_templates",&error);
  }
  else {
    filter=weed_instances[idx];
    num_channels=weed_leaf_num_elements(filter,"out_channels");
    if (num_channels>0) channels=weed_get_plantptr_array(filter,"out_channels",&error);
  }

  for (i=0;i<num_channels;i++) if (!weed_plant_has_leaf(channels[i],"disabled")||weed_get_boolean_value(channels[i],"disabled",&error)!=WEED_TRUE) enabled++;

  if (channels!=NULL) weed_free(channels);

  return enabled;
}



weed_plant_t *get_enabled_channel (weed_plant_t **channels, gint which) {
  // TODO - pass filter/instance
  //which starts at 0
  int i=0,error;
  if (channels==NULL) return NULL;
  while (which>-1) {
    if (!weed_plant_has_leaf(channels[i],"disabled")||weed_get_boolean_value(channels[i],"disabled",&error)==WEED_FALSE) which--;
    i++;
  }
  return channels[i-1];
}



/////////////////////////////////////////////////////////////////////



static gint check_for_lives(weed_plant_t *filter) {
  // for LiVES, currently:
  // all filters must take 0, 1 or 2 mandatory/optional inputs and provide 
  // 1 mandatory output or >1 optional outputs (for now)

  // all channels used must support a limited range of palettes (for now)

  gint chans_in_mand=0; // number of mandatory channels
  gint chans_in_opt_max=0; // number of usable (by LiVES) optional channels
  gint chans_out_mand=0;
  gint chans_out_opt_max=0;
  gboolean is_generator;
  
  int error,flags=0;
  int num_elements,i;
  weed_plant_t **array=NULL;

  // TODO - check seed types


  if (!weed_plant_has_leaf(filter,"name")) return 1;
  if (!weed_plant_has_leaf(filter,"author")) return 2;
  if (!weed_plant_has_leaf(filter,"version")) return 3;
  if (!weed_plant_has_leaf(filter,"process_func")) return 4;

  if (!weed_plant_has_leaf(filter,"flags")) weed_set_int_value(filter,"flags",0);
  else flags=weed_get_int_value(filter,"flags",&error);

  // for now we will only load realtime effects
  if (flags&WEED_FILTER_NON_REALTIME) return 5;

  // count number of mandatory and optional in_channels
  if (!weed_plant_has_leaf(filter,"in_channel_templates")) num_elements=0;
  else num_elements=weed_leaf_num_elements(filter,"in_channel_templates");

  if (num_elements>0) array=weed_get_plantptr_array(filter,"in_channel_templates",&error);

  for (i=0;i<num_elements;i++) {
#ifndef HAVE_POSIX_MEMALIGN
    if (weed_plant_has_leaf(array[i],"alignment")) return 12;
#endif
    if (!weed_plant_has_leaf(array[i],"name")||!weed_plant_has_leaf(array[i],"palette_list")) {
      weed_free(array);
      return 6;
    }

    if (!weed_plant_has_leaf(array[i],"flags")) weed_set_int_value(array[i],"flags",0);

    if (weed_plant_has_leaf(array[i],"optional")&&weed_get_boolean_value(array[i],"optional",&error)==WEED_TRUE) {
      // is optional
      if (check_weed_palette_list (array[i],WEED_PALETTE_BGR24)||check_weed_palette_list (array[i],WEED_PALETTE_RGBA32)||check_weed_palette_list(array[i],WEED_PALETTE_RGB24)||check_weed_palette_list(array[i],WEED_PALETTE_BGRA32)) chans_in_opt_max++;
      weed_set_boolean_value(array[i],"disabled",WEED_TRUE);
    }
    else {
      // is mandatory
      if (!check_weed_palette_list (array[i],WEED_PALETTE_BGR24)&&!check_weed_palette_list (array[i],WEED_PALETTE_RGBA32)&&!check_weed_palette_list(array[i],WEED_PALETTE_RGB24)&&!check_weed_palette_list(array[i],WEED_PALETTE_BGRA32)) {
#ifdef DEBUG_WEED
	g_printerr("in_channel template failed palette test\n");
#endif
	weed_free(array);
	return 7;
	
      }
      chans_in_mand++;
    }
  }
  if (num_elements>0) weed_free(array);
  if (chans_in_mand>2) return 8; // we dont handle mixers yet...

  is_generator=(chans_in_mand==0);

  // count number of mandatory and optional out_channels
  if (!weed_plant_has_leaf(filter,"out_channel_templates")) num_elements=0;
  else num_elements=weed_leaf_num_elements(filter,"out_channel_templates");

  if (num_elements>0) array=weed_get_plantptr_array(filter,"out_channel_templates",&error);

  // restrict it just to BGR24 and RGBA32, except for generators, which can return RGB24
  for (i=0;i<num_elements;i++) {
#ifndef HAVE_POSIX_MEMALIGN
    if (weed_plant_has_leaf(array[i],"alignment")) return 12;
#endif
    if (!weed_plant_has_leaf(array[i],"name")||!weed_plant_has_leaf(array[i],"palette_list")) {
      weed_free(array);
      return 9;
    }

    if (!weed_plant_has_leaf(array[i],"flags")) weed_set_int_value(array[i],"flags",0);

    if (weed_plant_has_leaf(array[i],"optional")&&weed_get_boolean_value(array[i],"optional",&error)==WEED_TRUE) {
      // is optional
      if (check_weed_palette_list (array[i],WEED_PALETTE_BGR24)||check_weed_palette_list (array[i],WEED_PALETTE_RGBA32)||check_weed_palette_list(array[i],WEED_PALETTE_RGB24)||check_weed_palette_list(array[i],WEED_PALETTE_BGRA32)) chans_out_opt_max++;
      weed_set_boolean_value(array[i],"disabled",WEED_TRUE);
    }
    else {
      // is mandatory
      if (!check_weed_palette_list (array[i],WEED_PALETTE_BGR24)&&!check_weed_palette_list (array[i],WEED_PALETTE_RGBA32)&&!check_weed_palette_list(array[i],WEED_PALETTE_RGB24)&&!check_weed_palette_list(array[i],WEED_PALETTE_BGRA32)) {
#ifdef DEBUG_WEED
	g_printerr("out_channel template failed palette test\n");
#endif
	weed_free(array);
	return 10;
      }
      chans_out_mand++;
    }
  }
  if (num_elements>0) weed_free(array);

  if (chans_out_mand>1||(chans_out_mand+chans_out_opt_max<1)) return 11;

  weed_add_plant_flags(filter,WEED_LEAF_READONLY_PLUGIN);
  if (weed_plant_has_leaf(filter,"gui")) {
    weed_plant_t *gui=weed_get_plantptr_value(filter,"gui",&error);
    weed_add_plant_flags(gui,WEED_LEAF_READONLY_PLUGIN);
  }

  return 0;
}



static gboolean set_in_channel_palettes (gint idx, gint num_channels) {
  // set in channel palettes for filter[idx]
  // we also enable optional channels if we have to
  // in this case we fill first the mandatory channels,
  // then if necessary the optional channels

  // we return FALSE if we could not satisfy the request
  weed_plant_t **chantmpls=NULL;
  int def_palette;

  int num_elements=weed_leaf_num_elements(weed_filters[idx],"in_channel_templates"),i,error;

  if (num_elements<num_channels) return FALSE;

  chantmpls=weed_get_plantptr_array(weed_filters[idx],"in_channel_templates",&error);
  // our start state is with all optional channels disabled

  // fill mandatory channels first
  for (i=0;i<num_elements;i++) {
    if (check_weed_palette_list (chantmpls[i],WEED_PALETTE_BGR24)) weed_set_int_value(chantmpls[i],"current_palette",WEED_PALETTE_BGR24);
    else weed_set_int_value(chantmpls[i],"current_palette",WEED_PALETTE_RGBA32);
    if (!weed_plant_has_leaf(chantmpls[i],"optional")) num_channels--; // mandatory channel
  }

  if (num_channels>0) {
    // OK we need to use some optional channels
    for (i=0;i<num_elements&&num_channels>0;i++) if (weed_plant_has_leaf(chantmpls[i],"optional")) {
      weed_set_boolean_value(chantmpls[i],"disabled",WEED_FALSE);
      num_channels--;
    }
  }
  if (num_channels>0) {
    if (chantmpls!=NULL) weed_free(chantmpls);
    return FALSE;
  }

  // now we set match channels
  def_palette=weed_get_int_value(chantmpls[0],"current_palette",&error);
  for (i=1;i<num_elements;i++) {
    int channel_flags=weed_get_int_value(chantmpls[i],"flags",&error);
    if (!(channel_flags&WEED_CHANNEL_PALETTE_CAN_VARY)) {
      if (check_weed_palette_list(chantmpls[i],def_palette)) weed_set_int_value(chantmpls[i],"current_palette",def_palette);
      else return FALSE;
    }
  }
  if (chantmpls!=NULL) weed_free(chantmpls);
  return TRUE;
}



static gboolean set_out_channel_palettes (gint idx, gint num_channels) {
  // set in channel palettes for filter[idx]
  // we also enable optional channels if we have to
  // in this case we fill first the mandatory channels,
  // then if necessary the optional channels

  // we return FALSE if we could not satisfy the request
  weed_plant_t **chantmpls=NULL;
  weed_plant_t **in_chantmpls=NULL;

  int num_elements=weed_leaf_num_elements(weed_filters[idx],"out_channel_templates"),i,error;
  int num_in_elements=weed_leaf_num_elements(weed_filters[idx],"in_channel_templates");

  int def_palette;

  if (num_elements<num_channels) return FALSE;

  chantmpls=weed_get_plantptr_array(weed_filters[idx],"out_channel_templates",&error);
  // our start state is with all optional channels disabled

  // fill mandatory channels first
  for (i=0;i<num_elements;i++) {
    if (check_weed_palette_list (chantmpls[i],WEED_PALETTE_BGR24)) weed_set_int_value(chantmpls[i],"current_palette",WEED_PALETTE_BGR24);
    else if (check_weed_palette_list (chantmpls[i],WEED_PALETTE_RGB24)) weed_set_int_value(chantmpls[i],"current_palette",WEED_PALETTE_RGB24);
    else weed_set_int_value(chantmpls[i],"current_palette",WEED_PALETTE_RGBA32);
    if (!weed_plant_has_leaf(chantmpls[i],"optional")) num_channels--; // mandatory channel
  }

  if (num_channels>0) {
    // OK we need to use some optional channels
    for (i=0;i<num_elements&&num_channels>0;i++) if (weed_plant_has_leaf(chantmpls[i],"optional")) {
      weed_set_boolean_value(chantmpls[i],"disabled",WEED_FALSE);
      num_channels--;
    }
  }

  if (num_channels>0) {
    if (chantmpls!=NULL) weed_free(chantmpls);
    return FALSE;
  }

  // now we set match channels
  if (num_in_elements) {
    in_chantmpls=weed_get_plantptr_array(weed_filters[idx],"in_channel_templates",&error);
    def_palette=weed_get_int_value(in_chantmpls[0],"current_palette",&error);
    weed_free(in_chantmpls);
  }
  else def_palette=weed_get_int_value(chantmpls[0],"current_palette",&error);
  for (i=0;i<num_elements;i++) {
    int channel_flags=weed_get_int_value(chantmpls[i],"flags",&error);
    if (!(channel_flags&WEED_CHANNEL_PALETTE_CAN_VARY)) {
      if (check_weed_palette_list(chantmpls[i],def_palette)) weed_set_int_value(chantmpls[i],"current_palette",def_palette);
      else return FALSE;
    }
  }
  if (chantmpls!=NULL) weed_free(chantmpls);
  return TRUE;
}




static void load_weed_plugin (gchar *plugin_name, gchar *plugin_path) {
  gchar *filter_name;
  weed_setup_f setup_fn;
  weed_bootstrap_f bootstrap=&weed_bootstrap_func;
  
  int error,reason,idx=num_weed_filters;

  weed_plant_t *plugin_info,**filters;
  void *handle;
  int filters_in_plugin;
  int mode=-1,kmode=0;
  int i;

  key++;

  // walk list and create fx structures

#ifdef DEBUG_WEED
  g_printerr("Checking plugin %s\n",plugin_path);
#endif

  if ((handle=dlopen(plugin_path,RTLD_NOW))) {
    setup_fn=dlsym(handle,"weed_setup");
    plugin_info=(*setup_fn)(bootstrap);
    if (plugin_info==NULL||(filters_in_plugin=check_weed_plugin_info(plugin_info))<1) {
      g_printerr ("error loading plugin %s\n",plugin_path);
      dlclose (handle);
      return;
    }
    weed_set_voidptr_value(plugin_info,"handle",handle);
    weed_set_string_value(plugin_info,"name",plugin_name); // for hashname
    
    weed_add_plant_flags(plugin_info,WEED_LEAF_READONLY_PLUGIN);
    
    filters=weed_get_plantptr_array(plugin_info,"filters",&error);
    
    while (idx<MAX_WEED_FILTERS&&mode<filters_in_plugin-1) {
      mode++;
      filter_name=weed_get_string_value(filters[mode],"name",&error);
      if (!(reason=check_for_lives(filters[mode]))) {
	gboolean dup=FALSE;
	weed_filters[idx]=filters[mode];
	num_weed_filters++;
	hashnames[idx]=make_weed_hashname(idx);
	for (i=0;i<idx;i++) {
	  if (!strcmp(hashnames[idx],hashnames[i])) {
	    // skip dups
	    g_free(hashnames[idx]);
	    hashnames[idx]=NULL;
	    weed_filters[idx]=NULL;
	    num_weed_filters--;
	    dup=TRUE;
	    break;
	  }
	}
	if (!dup) {
#ifdef DEBUG_WEED
	  if (key<FX_KEYS_PHYSICAL) d_print(g_strdup_printf("Loaded filter \"%s\" in plugin \"%s\"; assigned to key ctrl-%d, mode %d.\n",filter_name,plugin_name,key+1,kmode+1));
	  else d_print(g_strdup_printf("Loaded filter \"%s\" in plugin \"%s\", no key assigned\n",filter_name,plugin_name));
#endif
	
	  // we start with all optional channels disabled (unless forced to use them)
	  if (enabled_in_channels(idx,TRUE)>0) {
	    if (enabled_in_channels(idx,TRUE)==1) {
	      set_in_channel_palettes (idx,1);
	    }
	    else {
	      set_in_channel_palettes (idx,2);
	    }
	  }
	  set_out_channel_palettes (idx,1);
	  if (key<FX_KEYS_PHYSICAL) {
	    key_to_fx[key][kmode]=idx;
	  }
	  idx++;
	  kmode++;
	}
      }
#ifdef DEBUG_WEED
      else g_printerr(g_strdup_printf("Unsuitable filter \"%s\" in plugin \"%s\", reason code %d\n",filter_name,plugin_name,reason));
#endif
      weed_free(filter_name);
    }
    weed_free(filters);
  }
  else g_printerr(g_strdup_printf (_("Info: Unable to load plugin %s\n"),plugin_path));
  
  // TODO - add any rendered effects to fx submenu
}


void weed_load_all (void) {
// get list of plugins from directory and create our fx
  int i,j,plugin_idx,subdir_idx;

  GList *weed_plugin_list,*weed_plugin_sublist;

  gchar *lives_weed_plugin_path=g_strdup_printf("%s%s%s",prefs->prefix_dir,PLUGIN_DIR,PLUGIN_WEED_FX_BUILTIN),*weed_plugin_path,*weed_p_path;
  gchar *subdir_path,*subdir_name,*plugin_path,*plugin_name;
  gint numdirs;
  gchar **dirs;

  gint listlen;

  key=-1;

  num_weed_filters=0;

#ifdef DEBUG_WEED
  g_printerr("In weed init\n");
#endif

  // danger Will Robinson !
  fg_gen_to_start=fg_generator_key=fg_generator_clip=fg_generator_mode=-1;
  bg_gen_to_start=bg_generator_key=bg_generator_mode=-1;

  // reset all instances to NULL
  for (i=0;i<MAX_WEED_INSTANCES;i++) weed_instances[i]=NULL;

  for (i=0;i<FX_KEYS_MAX;i++) {
    key_modes[i]=0;  // current active mode of each key
    effect_map[i]=NULL; // maps effects in order of application for multitrack rendering
    for (j=0;j<MAX_MODES_PER_KEY;j++) {
      key_to_instance[i][j]=key_to_fx[i][j]=-1;
    }
  }
  effect_map[FX_KEYS_MAX+1]=NULL;
  for (i=0;i<FX_KEYS_MAX_VIRTUAL;i++) {
    init_events[i]=NULL;
  }

  next_free_key=FX_KEYS_MAX_VIRTUAL;

  for (i=0;i<MAX_WEED_FILTERS;i++) {
    weed_filters[i]=NULL; // TODO - ???
    hashnames[i]=NULL;
  }

  weed_p_path=getenv("WEED_PLUGIN_PATH");
  if (weed_p_path==NULL) weed_p_path=g_strdup("");
  weed_plugin_path=g_filename_to_utf8(weed_p_path,-1,NULL,NULL,NULL);
  g_free(weed_p_path);
  if (weed_plugin_path==NULL) weed_plugin_path=g_strdup("");
  if (strstr(weed_plugin_path,lives_weed_plugin_path)==NULL) {
    gchar *tmp=g_strconcat(strcmp(weed_plugin_path,"")?":":"",lives_weed_plugin_path,NULL);
    g_free(weed_plugin_path);
    weed_plugin_path=tmp;
    setenv("WEED_PLUGIN_PATH",weed_plugin_path,1);
  }
  g_free(lives_weed_plugin_path);

  // first we parse the weed_plugin_path
  numdirs=get_token_count(weed_plugin_path,':');
  dirs=g_strsplit(weed_plugin_path,":",-1);
  for (i=0;i<numdirs;i++) {
    // get list of all files
    weed_plugin_list=get_plugin_list(PLUGIN_EFFECTS_WEED,TRUE,dirs[i]);
    listlen=g_list_length(weed_plugin_list);
    
    // parse twice, first we get the plugins, then 1 level of subdirs
    for (plugin_idx=0;plugin_idx<listlen;plugin_idx++) {
      plugin_name=g_list_nth_data(weed_plugin_list,plugin_idx);
      if (strlen(plugin_name)<5||strcmp(".wo",plugin_name+strlen(plugin_name)-3)) {
	continue;
      }
      weed_plugin_list=g_list_remove(weed_plugin_list,plugin_name);
      listlen--;
      plugin_idx--;
      plugin_path=g_strdup_printf("%s/%s",dirs[i],plugin_name);
      load_weed_plugin(plugin_name,plugin_path);
      g_free(plugin_name);
      g_free(plugin_path);
    }
    
    // get 1 level of subdirs
    for (subdir_idx=0;subdir_idx<listlen;subdir_idx++) {
      subdir_name=g_list_nth_data(weed_plugin_list,subdir_idx);
      subdir_path=g_strdup_printf("%s/%s",dirs[i],subdir_name);
      if (!g_file_test(subdir_path, G_FILE_TEST_IS_DIR)||!strcmp(subdir_name,"icons")||!strcmp(subdir_name,"data")) {
	g_free(subdir_name);
	g_free(subdir_path);
	continue;
      }
      g_free(subdir_name);
      weed_plugin_sublist=get_plugin_list(PLUGIN_EFFECTS_WEED,TRUE,subdir_path);
      
      for (plugin_idx=0;plugin_idx<g_list_length(weed_plugin_sublist);plugin_idx++) {
	plugin_name=g_list_nth_data(weed_plugin_sublist,plugin_idx);
	if (strlen(plugin_name)<5||strcmp(".wo",plugin_name+strlen(plugin_name)-3)) {
	  g_free(plugin_name);
	  continue;
	}
	plugin_path=g_strdup_printf("%s/%s",subdir_path,plugin_name);
	load_weed_plugin(plugin_name,plugin_path);
	g_free(plugin_name);
	g_free(plugin_path);
      }
      if (weed_plugin_sublist!=NULL) g_list_free(weed_plugin_sublist);
      g_free(subdir_path);
    }
    if (weed_plugin_list!=NULL) g_list_free(weed_plugin_list);
  }
  g_strfreev(dirs);
  g_free(weed_plugin_path);

  d_print(g_strdup_printf(_ ("Succesfully loaded %d Weed filters\n"),num_weed_filters));

}



void weed_unload_all(void) {
  int i,error;
  weed_plant_t *filter,*plugin_info;
  void *handle;

  mainw->num_tr_applied=0;
  weed_deinit_all();
  for (i=0;i<num_weed_filters;i++) {
    filter=weed_filters[i];
    plugin_info=weed_get_plantptr_value(filter,"plugin_info",&error);
    handle=weed_get_voidptr_value(plugin_info,"handle",&error);
    if (handle!=NULL) {
      dlclose(handle);
      handle=NULL;
      weed_set_voidptr_value(plugin_info,"handle",handle);
    }
    // call weed_desetup()
    //weed_filter_free(filter); // etc...
  }
}


static void weed_channels_free (weed_plant_t *inst) {
  weed_plant_t **channels;
  int i,error;
  int num_channels;

  num_channels=weed_leaf_num_elements(inst,"in_channels");
  channels=weed_get_plantptr_array(inst,"in_channels",&error);
  for (i=0;i<num_channels;i++) weed_plant_free(channels[i]);
  weed_free(channels);

  num_channels=weed_leaf_num_elements(inst,"out_channels");
  channels=weed_get_plantptr_array(inst,"out_channels",&error);
  for (i=0;i<num_channels;i++) weed_plant_free(channels[i]);
  weed_free(channels);
}

static void weed_gui_free (weed_plant_t *plant) {
  weed_plant_t *gui;
  int error;

  if (weed_plant_has_leaf(plant,"gui")) {
    gui=weed_get_plantptr_value(plant,"gui",&error);
    weed_plant_free(gui);
  }
}


static void weed_parameters_free (weed_plant_t *inst) {
  weed_plant_t **parameters;
  int i,error;
  int num_parameters;

  num_parameters=weed_leaf_num_elements(inst,"in_parameters");
  parameters=weed_get_plantptr_array(inst,"in_parameters",&error);
  for (i=0;i<num_parameters;i++) {
    weed_gui_free(parameters[i]);
    weed_plant_free(parameters[i]);
  }
  weed_free(parameters);

  num_parameters=weed_leaf_num_elements(inst,"out_parameters");
  parameters=weed_get_plantptr_array(inst,"out_parameters",&error);
  for (i=0;i<num_parameters;i++) {
    weed_gui_free(parameters[i]);
    weed_plant_free(parameters[i]);
  }
  weed_free(parameters);
}



static void weed_free_instance (weed_plant_t *inst) {
  weed_channels_free(inst);
  weed_parameters_free(inst);
  weed_plant_free(inst);
}



void weed_generator_end (weed_plant_t *inst) {
  // generator has stopped for one of the following reasons:
  // efect was de-inited; clip (bg/fg) was changed; playback stopped with fg
  gboolean is_bg=FALSE;
  gint current_file=mainw->current_file;
  gboolean clip_switched=mainw->clip_switched;
  weed_plant_t *filter;
  weed_deinit_f filter_deinit;
  int error;

  if (inst==NULL) {
    g_printerr("  WARNING: inst was NULL !    ");
    return;
  }
  if (mainw->blend_file!=-1&&mainw->blend_file!=current_file&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->ext_src==inst) is_bg=TRUE;

  if (!is_bg&&mainw->whentostop==STOP_ON_VID_END&&mainw->playing_file>0) {
    // we will close the file after playback stops
    mainw->cancelled=CANCEL_GENERATOR_END;
    return;
  }
  
  if (rte_window!=NULL) {
    if (!is_bg) {
      rtew_set_param_button (fg_generator_key,fg_generator_mode,FALSE);
      rtew_set_keych(fg_generator_key,FALSE);
    }
    else {
      rtew_set_param_button (bg_generator_key,bg_generator_mode,FALSE);
      rtew_set_keych(bg_generator_key,FALSE);
    }
  }

  if (!is_bg) close_pwindow(fg_generator_key,fg_generator_mode,FALSE);
  else close_pwindow(bg_generator_key,bg_generator_mode,FALSE);

  filter=weed_get_plantptr_value(inst,"filter_class",&error);
  if (weed_plant_has_leaf(filter,"deinit_func")) {
    filter_deinit=weed_get_voidptr_value(filter,"deinit_func",&error);
    if (filter_deinit!=NULL) (*filter_deinit)(inst);
  }
  weed_free_instance(inst);

  current_file=mainw->current_file;

  if (is_bg) {
    weed_instances[key_to_instance[bg_generator_key][bg_generator_mode]]=NULL;
    key_to_instance[bg_generator_key][bg_generator_mode]=-1;
    if (mainw->rte&(GU641<<bg_generator_key)) mainw->rte^=(GU641<<bg_generator_key);
    bg_gen_to_start=bg_generator_key=bg_generator_mode=-1;
    mainw->current_file=mainw->blend_file;
    mainw->blend_file=-1;
  }
  else {
    weed_instances[key_to_instance[fg_generator_key][fg_generator_mode]]=NULL;
    key_to_instance[fg_generator_key][fg_generator_mode]=-1;
    if (mainw->rte&(GU641<<fg_generator_key)) mainw->rte^=(GU641<<fg_generator_key);
    fg_gen_to_start=fg_generator_key=fg_generator_clip=fg_generator_mode=-1;
    if (mainw->blend_file==mainw->current_file) mainw->blend_file=-1;
  }

  // TODO
  close_current_file();
  if (mainw->pre_src_file!=-1) {
    if (mainw->playing_file>-1) do_quick_switch (mainw->pre_src_file);
    else switch_to_file (mainw->current_file,mainw->pre_src_file);
  }
  if (mainw->current_file==current_file) mainw->clip_switched=clip_switched;
}


static weed_plant_t **weed_channels_create (weed_plant_t *filter, gboolean in) {
  weed_plant_t **channels,**chantmpls;
  int num_channels;
  int i,error;

  if (in) num_channels=weed_leaf_num_elements(filter,"in_channel_templates");
  else num_channels=weed_leaf_num_elements(filter,"out_channel_templates");

  if (num_channels==0) return NULL;
  if (in) chantmpls=weed_get_plantptr_array(filter,"in_channel_templates",&error);
  else chantmpls=weed_get_plantptr_array(filter,"out_channel_templates",&error);

  channels=g_malloc((num_channels+1)*sizeof(weed_plant_t *));

  for (i=0;i<num_channels;i++) {
    channels[i]=weed_plant_new(WEED_PLANT_CHANNEL);
    weed_set_plantptr_value(channels[i],"template",chantmpls[i]);
    weed_set_int_value(channels[i],"current_palette",weed_get_int_value(chantmpls[i],"current_palette",&error));
    if (weed_plant_has_leaf(chantmpls[i],"disabled")) (weed_set_boolean_value(channels[i],"disabled",weed_get_boolean_value(chantmpls[i],"disabled",&error)));
    weed_add_plant_flags(chantmpls[i],WEED_LEAF_READONLY_PLUGIN);
  }
  channels[num_channels]=NULL;
  weed_free(chantmpls);
  return channels;
}

static weed_plant_t **weed_params_create (weed_plant_t *filter, gboolean in) {
  // TODO - check all mandatory properties are present

  weed_plant_t **params,**paramtmpls;
  int num_params;
  int i,error;

  if (in) num_params=weed_leaf_num_elements(filter,"in_parameter_templates");
  else num_params=weed_leaf_num_elements(filter,"out_parameter_templates");

  if (num_params==0) return NULL;
  if (in) paramtmpls=weed_get_plantptr_array(filter,"in_parameter_templates",&error);
  else paramtmpls=weed_get_plantptr_array(filter,"out_parameter_templates",&error);

  params=g_malloc((num_params+1)*sizeof(weed_plant_t *));

  for (i=0;i<num_params;i++) {
    params[i]=weed_plant_new(WEED_PLANT_PARAMETER);
    weed_set_plantptr_value(params[i],"template",paramtmpls[i]);
    weed_leaf_copy(paramtmpls[i],"default",params[i],"value");
    weed_add_plant_flags(paramtmpls[i],WEED_LEAF_READONLY_PLUGIN);
  }
  params[num_params]=NULL;
  weed_free(paramtmpls);
  return params;
}



static void set_default_channel_sizes (weed_plant_t **in_channels, weed_plant_t **out_channels) {
  // set some reasonable default channel sizes when we first init the effect
  int error,i,j;
  weed_plant_t *channel;
  void **pixel_data;
  int numplanes,width,height;
  int *rowstrides,def_rowstride;
  gboolean is_gen=TRUE;

  for (i=0;in_channels!=NULL&&in_channels[i]!=NULL;i++) {
    is_gen=FALSE;
    channel=in_channels[i];
    width=height=4;
    weed_set_int_value(channel,"width",width);
    weed_set_int_value(channel,"height",height);
    def_rowstride=width*3;
    // try to set channel size first
    set_channel_size(channel,320,240,1,&def_rowstride);

    // create empty data for the palette and get the actual sizes
    create_empty_pixel_data(channel);
    width=weed_get_int_value(channel,"width",&error);
    height=weed_get_int_value(channel,"height",&error);
    numplanes=weed_leaf_num_elements(channel,"rowstrides");
    rowstrides=weed_get_int_array(channel,"rowstrides",&error);
    set_channel_size(channel,width,height,numplanes,rowstrides);
    weed_free(rowstrides);
    pixel_data=weed_get_voidptr_array(channel,"pixel_data",&error);
    for (j=0;j<numplanes;j++) g_free(pixel_data[j]);
    weed_free(pixel_data);
  }

  for (i=0;out_channels!=NULL&&out_channels[i]!=NULL;i++) {
    channel=out_channels[i];
    width=is_gen?DEF_GEN_WIDTH:320;
    height=is_gen?DEF_GEN_HEIGHT:240;
    weed_set_int_value(channel,"width",width);
    weed_set_int_value(channel,"height",height);
    def_rowstride=width*3;
    // try to set channel size first
    set_channel_size(channel,width,height,1,&def_rowstride);

    // create empty data for the palette and get the actual sizes
    create_empty_pixel_data(channel);
    width=weed_get_int_value(channel,"width",&error);
    height=weed_get_int_value(channel,"height",&error);
    numplanes=weed_leaf_num_elements(channel,"rowstrides");
    rowstrides=weed_get_int_array(channel,"rowstrides",&error);
    set_channel_size(channel,width,height,numplanes,rowstrides);
    weed_free(rowstrides);
    pixel_data=weed_get_voidptr_array(channel,"pixel_data",&error);
    for (j=0;j<numplanes;j++) g_free(pixel_data[j]);
    weed_free(pixel_data);
  }


}

static inline int array_count(weed_plant_t **array, gboolean set_readonly) {
  int i=0;
  while (array[i]!=NULL) {
    if (set_readonly) weed_add_plant_flags(array[i],WEED_LEAF_READONLY_PLUGIN);
    i++;
  }
  return i;
}


static weed_plant_t *weed_create_instance (weed_plant_t *filter, weed_plant_t **inc, weed_plant_t **outc, weed_plant_t **inp, weed_plant_t **outp) {
  // here we create a new filter_instance from the ingredients: filter_class, in_channels, out_channels, in_parameters, out_parameters
  weed_plant_t *inst=weed_plant_new(WEED_PLANT_FILTER_INSTANCE);
  weed_set_plantptr_value(inst,"filter_class",filter);
  if (inc!=NULL) weed_set_plantptr_array(inst,"in_channels",array_count(inc,TRUE),inc);
  if (outc!=NULL) weed_set_plantptr_array(inst,"out_channels",array_count(outc,TRUE),outc);
  if (inp!=NULL) weed_set_plantptr_array(inst,"in_parameters",array_count(inp,TRUE),inp);
  if (outp!=NULL) weed_set_plantptr_array(inst,"out_parameters",array_count(outp,TRUE),outp);

  weed_add_plant_flags(inst,WEED_LEAF_READONLY_PLUGIN);
  return inst;
}


static inline int get_next_free_instance(void) {
  int i;
  for (i=next_free_instance+1;i!=next_free_instance&&!(next_free_instance==-1&&i==MAX_WEED_INSTANCES);i++) if (weed_instances[(i=(i==MAX_WEED_INSTANCES?0:i))]==NULL) return i;
  return -1;
}


gboolean weed_init_effect(int hotkey) {
  // mainw->osc_block should be set to TRUE before calling this function !
  weed_plant_t *filter;
  weed_plant_t *new_instance;
  weed_plant_t **inc,**outc,**inp,**outp;
  gint num_tr_applied;
  gboolean fg_modeswitch=FALSE,is_trans=FALSE,gen_start=FALSE,is_modeswitch=FALSE;
  gint rte_keys=mainw->rte_keys;
  gint inc_count;
  event_list_t *event_list;
  weed_init_f filter_init;
  int error;

  if (hotkey<0) {
    is_modeswitch=TRUE;
    hotkey=-hotkey-1;
  }

  if (hotkey==fg_generator_key) {
    fg_modeswitch=TRUE;
  }

  if (!rte_key_valid (hotkey)) {
    return FALSE;
  }

  if (next_free_instance==-1) {
    d_print(_ ("Warning - no more effect slots left !\n"));
    return FALSE;
  }

  inc_count=enabled_in_channels(key_to_fx[hotkey][key_modes[hotkey]],TRUE);

  // TODO - block template channel changes
  // we must stop any old generators
  // warning = in future, bg *and* fg generators could change
  if (mainw->num_tr_applied&&mainw->blend_file!=-1&&mainw->blend_file!=mainw->current_file&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->clip_type==CLIP_TYPE_GENERATOR&&inc_count==0) {
    if (bg_gen_to_start==-1) {
      weed_generator_end(mainw->files[mainw->blend_file]->ext_src);
    }
    bg_gen_to_start=bg_generator_key=bg_generator_mode=-1;
    mainw->blend_file=-1;
  }
  if (mainw->current_file>0&&cfile->clip_type==CLIP_TYPE_GENERATOR&&(fg_modeswitch||inc_count==0)) {
    if (mainw->noswitch||mainw->is_processing||mainw->preview) return FALSE; // stopping fg gen will cause clip to switch
    if (mainw->playing_file>-1&&mainw->whentostop==STOP_ON_VID_END&&inc_count!=0) {
      mainw->cancelled=CANCEL_GENERATOR_END;
    }
    else {
      if (inc_count==0&&mainw->whentostop==STOP_ON_VID_END) mainw->whentostop=NEVER_STOP;
      weed_generator_end (cfile->ext_src);
      mainw->reinit_effects=TRUE;
      fg_generator_key=fg_generator_clip=fg_generator_mode=-1;
    }
  }

  filter=weed_filters[key_to_fx[hotkey][key_modes[hotkey]]];

  if (inc_count==2) {
    mainw->num_tr_applied++; // increase trans count
    if (mainw->num_tr_applied==1&&!is_modeswitch) {
      mainw->blend_file=mainw->current_file;
      mainw->blend_file_step=(mainw->current_file>-1&&cfile->pb_fps<0)?-1:1;
    }
  }
  else if (inc_count==0) {
    // aha - a generator
     if (mainw->playing_file==-1) {
      // if we are not playing, we will postpone creating the instance
      // this is a workaround for a problem in libvisual
      if (mainw->num_tr_applied&&mainw->current_file!=-1) {
	bg_gen_to_start=hotkey;
      }
      else fg_gen_to_start=hotkey;
      gen_start=TRUE;
    }
     else if (!fg_modeswitch&&!mainw->num_tr_applied&&(mainw->noswitch||mainw->is_processing||mainw->preview)) return FALSE;
  }

  // create channels for the instance
  filter=weed_filters[key_to_fx[hotkey][key_modes[hotkey]]];
  inc=weed_channels_create (filter,TRUE);
  outc=weed_channels_create (filter,FALSE);

  set_default_channel_sizes (inc,outc); // we set the initial channel sizes to some reasonable defaults

  inp=weed_params_create(filter,TRUE);
  outp=weed_params_create(filter,FALSE);

  new_instance=weed_create_instance(filter,inc,outc,inp,outp);
  // TODO - unblock template channel changes
  
  if (inc!=NULL) weed_free(inc);
  if (outc!=NULL) weed_free(outc);
  if (inp!=NULL) weed_free(inp);
  if (outp!=NULL) weed_free(outp);

  weed_instances[next_free_instance]=new_instance;
  key_to_instance[hotkey][key_modes[hotkey]]=next_free_instance;

  if (!gen_start) {
    if (weed_plant_has_leaf(filter,"init_func")) {
      filter_init=weed_get_voidptr_value(filter,"init_func",&error);
      if (filter_init!=NULL&&(error=(*filter_init)(new_instance))!=WEED_NO_ERROR) {
	gint weed_error;
	gchar *filter_name=weed_get_string_value(filter,"name",&weed_error);
	d_print (g_strdup_printf (_ ("Failed to start instance %s, error code %d\n"),filter_name,error));
	weed_free(filter_name);
	weed_free_instance(new_instance);
	weed_instances[key_to_instance[hotkey][key_modes[hotkey]]]=NULL;
	key_to_instance[hotkey][key_modes[hotkey]]=-1;
	if (is_trans) mainw->num_tr_applied--;
	return FALSE;
      }
    }
    set_param_gui_readonly(new_instance);
  }

  if (inc_count==0) {
    // generator start
    if (mainw->num_tr_applied&&!fg_modeswitch&&mainw->current_file>-1) {
      // transition is on, make into bg clip
      bg_generator_key=hotkey;
      bg_generator_mode=key_modes[hotkey];
    }
    else {
      // no transition, make fg (or kb was grabbed for fg generator)
      fg_generator_key=hotkey;
      fg_generator_mode=key_modes[hotkey];
    }

    // start the generator and maybe start playing
    num_tr_applied=mainw->num_tr_applied;
    if (fg_modeswitch) mainw->num_tr_applied=0; // force to fg

    // TODO - be more descriptive with error
    if (!weed_generator_start(new_instance)) {
      int weed_error;
      gchar *filter_name=weed_get_string_value(filter,"name",&weed_error);
      d_print (g_strdup_printf (_ ("Unable to start generator %s\n"),filter_name));
      weed_free(filter_name);
      if (mainw->num_tr_applied&&mainw->current_file>-1) {
	bg_gen_to_start=bg_generator_key=bg_generator_mode=-1;
      }
      else {
	fg_generator_key=fg_generator_clip=fg_generator_mode=-1;
      }
      if (fg_modeswitch) mainw->num_tr_applied=num_tr_applied;
      return FALSE;
    }
    // TODO - problem if modeswitch triggers playback
    if (fg_modeswitch) mainw->num_tr_applied=num_tr_applied;
    if (fg_generator_key!=-1) {
      mainw->rte|=(GU641<<fg_generator_key);
      mainw->clip_switched=TRUE;
    }
    if (bg_generator_key!=-1) {
      mainw->rte|=(GU641<<bg_generator_key);
      if (rte_window!=NULL) rtew_set_keych(bg_generator_key,TRUE);
    }
  }

  if (rte_keys==hotkey) {
    mainw->rte_keys=rte_keys;
    mainw->blend_factor=weed_get_blend_factor(rte_keys);
  }
  if (rte_window!=NULL) rtew_set_param_button (hotkey,key_modes[hotkey],TRUE);

  next_free_instance=get_next_free_instance();

  if (mainw->record&&mainw->playing_file>-1&&(prefs->rec_opts&REC_EFFECTS)&&inc_count>0) {
    // place this synchronous with the preceding frame
    event_list=append_effect_init_event (mainw->event_list,mainw->currticks-mainw->origticks,key_to_fx[hotkey][key_modes[hotkey]]);
    if (mainw->event_list==NULL) mainw->event_list=event_list;
    init_events[hotkey]=mainw->event_list->last;
    create_effect_map(); // we create effect_map event_t * array with ordered effects
    add_effect_map_event (mainw->event_list,mainw->currticks-mainw->origticks,effect_map);
  }

  return TRUE;
}



void weed_deinit_effect(int hotkey) {
  // mainw->osc_block should be set before calling this function !
  // caller should also handle mainw->rte

  weed_plant_t *filter;
  weed_plant_t *instance;
  gboolean is_modeswitch=FALSE;
  gboolean was_transition=FALSE;
  gint num_in_chans;
  gint idx;
  weed_deinit_f filter_deinit;
  int error;

  if (hotkey<0) {
    is_modeswitch=TRUE;
    hotkey=-hotkey-1;
  }

  if (hotkey>FX_KEYS_MAX) return;

  if ((idx=key_to_instance[hotkey][key_modes[hotkey]])==-1) return;

  instance=weed_instances[idx];
  if (instance==NULL) return;

  num_in_chans=enabled_in_channels(idx,FALSE);

  if (num_in_chans==0) {
    // is generator
    if (mainw->playing_file>-1&&mainw->whentostop==STOP_ON_VID_END) {
      mainw->cancelled=CANCEL_GENERATOR_END;
    }
    else {
      weed_generator_end (instance);
    }
    return;
  }
  else {
    if (rte_window!=NULL) rtew_set_param_button (hotkey,key_modes[hotkey],FALSE);
    close_pwindow(hotkey,key_modes[hotkey],FALSE);
    filter=weed_get_plantptr_value(instance,"filter_class",&error);
    if (weed_plant_has_leaf(filter,"deinit_func")) {
      filter_deinit=weed_get_voidptr_value(filter,"deinit_func",&error);
      if (filter_deinit!=NULL) (*filter_deinit)(instance);
    }
  }

  if (num_in_chans==2) {
    was_transition=TRUE;
    mainw->num_tr_applied--;
  }

  weed_free_instance(instance);

  weed_instances[idx]=NULL;
  key_to_instance[hotkey][key_modes[hotkey]]=-1;

  if (was_transition&&!is_modeswitch) {
    if (mainw->num_tr_applied<1) {
      if (bg_gen_to_start!=-1) bg_gen_to_start=-1;
      if (mainw->blend_file!=-1&&mainw->blend_file!=mainw->current_file&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->clip_type==CLIP_TYPE_GENERATOR) {
	// all transitions off, so end the bg generator
	weed_generator_end (mainw->files[mainw->blend_file]->ext_src);
      }
      mainw->blend_file=-1;
    }
  }
  if (mainw->record&&mainw->playing_file>-1&&init_events[hotkey]!=NULL&&(prefs->rec_opts&REC_EFFECTS)&&num_in_chans>0) {
    // place this synchronous with the preceding frame
    add_effect_deinit_event (mainw->event_list,mainw->currticks-mainw->origticks,init_events[hotkey]);
    init_events[hotkey]=NULL;
    create_effect_map(); // we create effect_map event_t * array with ordered effects
    add_effect_map_event (mainw->event_list,mainw->currticks-mainw->origticks,effect_map);
  }
}


void deinit_render_effects (void) {
  // during rendering/render preview, we use the "keys" 64 -> FX_KEYS_MAX
  // here we deinit all active ones, similar to weed_deinit_all, but we use higher keys
  int i;

  for (i=64;i<FX_KEYS_MAX;i++) {
    if (key_to_instance[i][0]!=-1) {
      weed_deinit_effect (i);
    }
  }
}


void weed_deinit_all(void) {
  // deinit all except generators
  // this is called on ctrl-0 or on shutdown
  int i;
  gint idx;

  mainw->osc_block=TRUE;
  if (mainw->playing_file==-1&&rte_window!=NULL) {
    rtew_set_keygr(-1);
    mainw->rte_keys=-1;
    mainw->last_grabable_effect=-1;
  }

  for (i=0;i<FX_KEYS_PHYSICAL&&rte_key_valid (i)&&i<num_weed_filters;i++) {
    if (mainw->playing_file==-1&&rte_window!=NULL) rtew_set_keych(i,FALSE);
    if ((mainw->rte&(GU641<<i))) {
      if ((idx=key_to_instance[i][key_modes[i]])!=-1&&weed_instances[idx]!=NULL) {
	if (enabled_in_channels(idx,FALSE)>0) {
	  mainw->noswitch=TRUE;
	  weed_deinit_effect(i);
	  if (mainw->rte&(GU641<<i)) mainw->rte^=(GU641<<i);
	  if (rte_window!=NULL) {
	    rtew_set_keych(i,FALSE);
	    if (i==mainw->rte_keys) {
	      rtew_set_keygr(-1);
	      mainw->rte_keys=-1;
	    }
	  }
	  if (i==mainw->last_grabable_effect) mainw->last_grabable_effect=-1;
	}
      }
    }
  }
  mainw->osc_block=FALSE;
}


/////////////////////
// special handling for generators (sources)


GdkPixbuf *gdk_pixbuf_new_from_generator (weed_plant_t *inst, weed_timecode_t tc) {
  // TODO - change to weed_layer_from_generator
  GdkPixbuf *pixbuf;
  weed_plant_t *channel,**out_channels;
  int num_channels;
  weed_plant_t *filter;
  weed_process_f filter_process;
  int error;
  int palette;
  weed_plant_t *chantmpl;

  if (inst==NULL) return NULL;
  if ((num_channels=weed_leaf_num_elements(inst,"out_channels"))==0) return NULL;
  out_channels=weed_get_plantptr_array(inst,"out_channels",&error);
  if ((channel=get_enabled_channel(out_channels,0))==NULL) {
    weed_free(out_channels);
    return NULL;
  }
  weed_free(out_channels);

  chantmpl=weed_get_plantptr_value(channel,"template",&error);
  palette=weed_get_int_value(chantmpl,"current_palette",&error);
  weed_set_int_value(channel,"current_palette",palette);
  
  create_empty_pixel_data(channel);

  // align memory if necessary
  if (weed_plant_has_leaf(chantmpl,"alignment")) {
    int alignment=weed_get_int_value(chantmpl,"alignment",&error);
    void **pixel_data=weed_get_voidptr_array(channel,"pixel_data",&error);
    int numplanes=weed_leaf_num_elements(channel,"pixel_data");
    int height=weed_get_int_value(channel,"height",&error);
    int *rowstrides=weed_get_int_array(channel,"rowstrides",&error);
    align(pixel_data,alignment,numplanes,height,rowstrides);
    weed_set_voidptr_array(channel,"pixel_data",numplanes,pixel_data);
    weed_free(pixel_data);
    weed_free(rowstrides);
  }

  weed_set_double_value(inst,"fps",cfile->pb_fps);

  filter=weed_get_plantptr_value(inst,"filter_class",&error);
  filter_process=weed_get_voidptr_value(filter,"process_func",&error);
  (*filter_process)(inst,tc);

  convert_layer_palette(channel,WEED_PALETTE_BGR24); // TODO - don't !
  pixbuf=layer_to_pixbuf(channel);

  return pixbuf;
}


gboolean weed_generator_start (weed_plant_t *inst) {
  // make an "ephemeral clip"

  // cf. yuv4mpeg.c
  // start "playing" but receive frames from a plugin

  // TODO - if filter has a "target_fps", and it is the foreground, we try to synch with it

  weed_plant_t **out_channels,*channel,*filter;
  gchar *filter_name;
  int error,num_channels;
  gint old_file=mainw->current_file;

  // create a virtual clip
  gint new_file=mainw->first_free_file;

  if (new_file==-1) {
    too_many_files();
    return FALSE;
  }

  if (mainw->current_file>-1&&cfile!=NULL&&cfile->clip_type==CLIP_TYPE_GENERATOR) {
    close_current_file();
    old_file=-1;
  }

  if (mainw->blend_file!=-1&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->clip_type==CLIP_TYPE_GENERATOR) return FALSE;

  if (mainw->is_processing&&!mainw->preview) return FALSE;
  if ((mainw->preview||(mainw->current_file>-1&&cfile!=NULL&&cfile->opening))&&(!mainw->num_tr_applied||mainw->blend_file==-1||mainw->blend_file==mainw->current_file)) return FALSE;

  if (mainw->playing_file==-1) mainw->pre_src_file=mainw->current_file;
  mainw->current_file=new_file;

  cfile=(file *)(g_malloc(sizeof(file)));
  create_cfile();
  cfile->clip_type=CLIP_TYPE_GENERATOR;
  get_next_free_file();

  filter=weed_get_plantptr_value(inst,"filter_class",&error);
  filter_name=weed_get_string_value(filter,"name",&error);
  g_snprintf(cfile->type,40,"generator:%s",filter_name);
  weed_free(filter_name);
  cfile->achans=0;
  cfile->asampsize=0;
  g_snprintf (cfile->handle,256,"ephemeral%d",mainw->current_file);

  // open as a clip with 1 frame
  cfile->start=cfile->end=cfile->frames=1;
  cfile->arps=cfile->arate=0;
  cfile->changed=FALSE;

  cfile->ext_src=inst;
 
  if (!mainw->num_tr_applied||old_file==-1) fg_generator_clip=new_file;

  cfile->pb_fps=cfile->fps=prefs->default_fps;

  if ((num_channels=weed_leaf_num_elements(inst,"out_channels"))==0) return FALSE;
  out_channels=weed_get_plantptr_array(inst,"out_channels",&error);
  if ((channel=get_enabled_channel(out_channels,0))==NULL) {
    weed_free(out_channels);
    return FALSE;
  }
  weed_free(out_channels);

  cfile->hsize=weed_get_int_value(channel,"width",&error);
  cfile->vsize=weed_get_int_value(channel,"height",&error);

  cfile->opening=FALSE;
  cfile->proc_ptr=NULL;

  // allow clip switching
  cfile->is_loaded=TRUE;

  // if not playing, start playing
  if (mainw->playing_file==-1) {
    if (!mainw->num_tr_applied||old_file==-1) {
      switch_to_file((mainw->current_file=old_file),new_file);
      set_main_title(cfile->file_name,0);
      mainw->play_start=1;
      mainw->play_end=INT_MAX;
      if (mainw->num_tr_applied) {
	mainw->blend_file=mainw->current_file;
      }
    }
    else {
      mainw->blend_file=mainw->current_file;
      mainw->current_file=old_file;
      mainw->play_start=cfile->start;
      mainw->play_end=cfile->end;
      mainw->playing_sel=FALSE;
    }
    if (mainw->play_window!=NULL&&old_file==-1) {
      // usually preview or load_preview_frame would do this
      g_signal_handler_block(mainw->play_window,mainw->pw_exp_func);
    }

    play_file();

    if (mainw->current_file>0&&(mainw->pre_src_file==-1||mainw->multitrack!=NULL)&&mainw->current_file!=mainw->scrap_file&&mainw->current_file!=new_file) mainw->pre_src_file=mainw->current_file;

    if (fg_generator_clip!=-1) {
      if (rte_window!=NULL) rtew_set_keych(fg_generator_key,FALSE);
      inst=mainw->files[fg_generator_clip]->ext_src;
      if (mainw->blend_file==-1||mainw->blend_file==mainw->current_file||mainw->blend_file!=new_file) {
	mainw->current_file=fg_generator_clip;
	weed_generator_end (inst);
      }
    }
    if (mainw->pre_src_file>-1&&mainw->multitrack==NULL) {
      switch_to_file ((mainw->current_file=0),mainw->pre_src_file);
      mainw->pre_src_file=-2;
    }
  }
  else {
    // already playing
    mainw->pre_src_file=mainw->current_file=old_file;
    if (mainw->cancelled==CANCEL_GENERATOR_END) mainw->cancelled=0;
    if (!mainw->num_tr_applied||old_file==-1) {
      if (mainw->current_file==-1) mainw->current_file=new_file;
      do_quick_switch (new_file);
      if (old_file==-1) mainw->whentostop=STOP_ON_VID_END;
    }
    else {
      mainw->blend_file=new_file;
    }
  }
  return TRUE;
}



void weed_bg_generator_end (weed_plant_t *inst) {
  // when we stop with a bg generator we want it to be restarted next time
  // i.e we will need a new clip for it
  int bg_gen_key=bg_generator_key;
  weed_generator_end (inst);
  bg_gen_to_start=bg_gen_key;
}



gboolean weed_playback_gen_start (void) {
  // init generators on pb. We have to do this after audio startup
  weed_plant_t *inst=NULL,*filter;
  int error=WEED_NO_ERROR;
  int weed_error;
  gchar *filter_name;

  if (mainw->is_rendering) return TRUE;

  mainw->osc_block=TRUE;
  if (bg_gen_to_start!=-1) {
    if (fg_gen_to_start==-1) {
      if (enabled_in_channels(key_to_fx[bg_gen_to_start][key_modes[bg_gen_to_start]],TRUE)==0) { // check is still gen
	inst=weed_instances[key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]];

	if (inst==NULL) {
	  // 2nd playback
	  gint playing_file=mainw->playing_file;
	  mainw->playing_file=-100; //kludge to stop playing a second time
	  if (!weed_init_effect (bg_gen_to_start)) {
	    error++;
	  }
	  mainw->playing_file=playing_file;
	  inst=weed_instances[key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]];
	}
	else {
	  filter=weed_get_plantptr_value(inst,"filter_class",&weed_error);
	  if (weed_plant_has_leaf(filter,"init_func")) {
	    filter=weed_get_plantptr_value(inst,"filter_class",&weed_error);
	    weed_init_f init_func=weed_get_voidptr_value(filter,"init_func",&error);
	    if (init_func!=NULL) error=(*init_func)(inst);
	  }
	}

	if (error!=WEED_NO_ERROR) {
	  if (inst!=NULL) {
	    filter=weed_get_plantptr_value(inst,"filter_class",&weed_error);
	    filter_name=weed_get_string_value(filter,"name",&weed_error);
	    d_print (g_strdup_printf (_ ("Failed to start generator %s\n"),filter_name));
	    weed_free(filter_name);
	    if (weed_plant_has_leaf(filter,"deinit_func")) {
	      weed_init_f deinit_func=weed_get_voidptr_value(filter,"deinit_func",&weed_error);
	      if (deinit_func!=NULL) (*deinit_func)(inst);
	    }
	    weed_free_instance(inst);
	  }
	  weed_instances[key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]]=NULL;
	  key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]=-1;
	  bg_gen_to_start=-1;
	  if (mainw->rte&(GU641<<bg_gen_to_start)) mainw->rte^=(GU641<<bg_gen_to_start);
	  mainw->osc_block=FALSE;
	  return FALSE;
	}
	mainw->files[mainw->blend_file]->ext_src=inst;
	next_free_instance=get_next_free_instance();
	if (rte_window!=NULL) rtew_set_param_button (bg_gen_to_start,key_modes[bg_gen_to_start],TRUE);
      }
    }
    bg_gen_to_start=-1;
  }

  if (fg_gen_to_start!=-1) {
    if (enabled_in_channels(key_to_fx[fg_gen_to_start][key_modes[fg_gen_to_start]],TRUE)==0) { // check is still gen
      inst=weed_instances[key_to_instance[fg_gen_to_start][key_modes[fg_gen_to_start]]];
      if (inst!=NULL) {
	filter=weed_get_plantptr_value(inst,"filter_class",&weed_error);
	if (weed_plant_has_leaf(filter,"init_func")) {
	  weed_init_f init_func=weed_get_voidptr_value(filter,"init_func",&weed_error);
	  if (init_func!=NULL) error=(*init_func)(inst);
	}
      }
      if (error!=WEED_NO_ERROR) {
	if (inst!=NULL) {
	  filter=weed_get_plantptr_value(inst,"filter_class",&weed_error);
	  filter_name=weed_get_string_value(filter,"name",&weed_error);
	  d_print (g_strdup_printf (_ ("Failed to start generator %s\n"),filter_name));
	  weed_free(filter_name);
	  if (weed_plant_has_leaf(filter,"deinit_func")) {
	    weed_init_f deinit_func=weed_get_voidptr_value(filter,"deinit_func",&weed_error);
	    if (deinit_func!=NULL) (*deinit_func)(inst);
	  }
	  weed_free_instance(inst);
	}
	weed_instances[key_to_instance[fg_gen_to_start][key_modes[fg_gen_to_start]]]=NULL;
	key_to_instance[fg_gen_to_start][key_modes[fg_gen_to_start]]=-1;
	fg_gen_to_start=-1;
	cfile->ext_src=NULL;
	mainw->osc_block=FALSE;
	return FALSE;
      }
      mainw->clip_switched=TRUE;
      cfile->ext_src=inst;
      next_free_instance=get_next_free_instance();
      if (rte_window!=NULL) rtew_set_param_button (fg_gen_to_start,key_modes[fg_gen_to_start],TRUE);
    }
    fg_gen_to_start=-1;
  }

  if (inst!=NULL) set_param_gui_readonly(inst);

  mainw->osc_block=FALSE;
  return TRUE;
}
 
//////////////////////////////////////////////////////////////////////////////
// weed parameter functions

weed_plant_t *weed_inst_in_param (weed_plant_t *inst, int param_num) {
  weed_plant_t **in_params,*param=NULL;
  int error,num_params;

  if (!weed_plant_has_leaf(inst,"in_parameters")) return NULL; // has no in_parameters

  num_params=weed_leaf_num_elements(inst,"in_parameters");
  if (num_params<=param_num) return NULL; // invalid parameter number

  in_params=weed_get_plantptr_array(inst,"in_parameters",&error);
  param=in_params[param_num];
  weed_free(in_params);
  return param;
}

char *get_weed_display_string (weed_plant_t *inst, int pnum) {
  // TODO - for setting defaults, we will need to create params
  char *disp_string;
  weed_plant_t *param=weed_inst_in_param(inst,pnum);
  weed_plant_t *ptmpl,*gui;
  int error;
  weed_display_f display_func;

  if (param==NULL) return NULL;

  ptmpl=weed_get_plantptr_value(param,"template",&error);
  if (!weed_plant_has_leaf(ptmpl,"gui")) return NULL;
  gui=weed_get_plantptr_value(ptmpl,"gui",&error);
  if (!weed_plant_has_leaf(gui,"display_func")) return NULL;

  display_func=weed_get_voidptr_value(gui,"display_func",&error);

  weed_leaf_set_flags(gui,"display_value",(weed_leaf_get_flags(gui,"display_value")|WEED_LEAF_READONLY_PLUGIN)^WEED_LEAF_READONLY_PLUGIN);
  (*display_func)(param);
  weed_leaf_set_flags(gui,"display_value",(weed_leaf_get_flags(gui,"display_value")|WEED_LEAF_READONLY_PLUGIN));

  if (!weed_plant_has_leaf(gui,"display_value")) return NULL;
  if (weed_leaf_seed_type(gui,"display_value")!=WEED_SEED_STRING) return NULL;
  disp_string=weed_get_string_value(gui,"display_value",&error);

  return disp_string;
}



#define KEYSCALE 255.


void weed_set_blend_factor(int hotkey) {
  // mainw->osc_block should be set to TRUE before calling this function !
  weed_plant_t *inst,*in_param,*paramtmpl;
  int idx=key_to_instance[hotkey][key_modes[hotkey]],error;
  int vali,mini,maxi;
  double vald,mind,maxd;
  GList *list=NULL;
  int weed_hint;

  if (hotkey<0) return;
  idx=key_to_instance[hotkey][key_modes[hotkey]];

  if (idx==-1||(inst=weed_instances[idx])==NULL) return;


  in_param=weed_inst_in_param(inst,0); // TODO ** - skip hidden params
  if (in_param==NULL) return;

  paramtmpl=weed_get_plantptr_value(in_param,"template",&error);
  weed_hint=weed_get_int_value(paramtmpl,"hint",&error);

  switch (weed_hint) {
  case WEED_HINT_INTEGER:
    vali=weed_get_int_value(in_param,"value",&error);
    mini=weed_get_int_value(paramtmpl,"min",&error);
    maxi=weed_get_int_value(paramtmpl,"max",&error);

    weed_set_int_value (in_param,"value",(int)((gdouble)mini+(mainw->blend_factor/KEYSCALE*(gdouble)(maxi-mini))+.5));
    vali=weed_get_int_value (in_param,"value",&error);

    list=g_list_append(list,g_strdup_printf("%d",vali));
    list=g_list_append(list,g_strdup_printf("%d",mini));
    list=g_list_append(list,g_strdup_printf("%d",maxi));
    update_pwindow(hotkey,0,list);
    g_list_free(list);

    break;
  case WEED_HINT_FLOAT:
    vald=weed_get_double_value(in_param,"value",&error);
    mind=weed_get_double_value(paramtmpl,"min",&error);
    maxd=weed_get_double_value(paramtmpl,"max",&error);

    weed_set_double_value (in_param,"value",mind+(mainw->blend_factor/KEYSCALE*(maxd-mind)));
    vald=weed_get_double_value (in_param,"value",&error);

    list=g_list_append(list,g_strdup_printf("%.4f",vald));
    list=g_list_append(list,g_strdup_printf("%.4f",mind));
    list=g_list_append(list,g_strdup_printf("%.4f",maxd));
    update_pwindow(hotkey,0,list);
    g_list_free(list);

    break;
  case WEED_HINT_SWITCH:
    vali=!!(int)mainw->blend_factor;
    weed_set_boolean_value (in_param,"value",vali);
    vali=weed_get_boolean_value (in_param,"value",&error);
    mainw->blend_factor=(gdouble)vali;

    list=g_list_append(list,g_strdup_printf("%d",vali));
    update_pwindow(hotkey,0,list);
    g_list_free(list);

    break;
  }
}



gint weed_get_blend_factor(int hotkey) {
  // mainw->osc_block should be set to TRUE before calling this function !

  weed_plant_t *inst,*in_param,*paramtmpl;
  int idx,error;
  int vali,mini,maxi;
  double vald,mind,maxd;
  int weed_hint;

  if (hotkey<0) return 0;
  idx=key_to_instance[hotkey][key_modes[hotkey]];

  if (idx==-1||(inst=weed_instances[idx])==NULL) return 0;

  if (!weed_plant_has_leaf(inst,"in_parameters")) return 0;
  in_param=weed_get_plantptr_value(inst,"in_parameters",&error);
  if (in_param==NULL) return 0;

  paramtmpl=weed_get_plantptr_value(in_param,"template",&error);
  weed_hint=weed_get_int_value(paramtmpl,"hint",&error);

  switch (weed_hint) {
  case WEED_HINT_INTEGER:
    vali=weed_get_int_value(in_param,"value",&error);
    mini=weed_get_int_value(paramtmpl,"min",&error);
    maxi=weed_get_int_value(paramtmpl,"max",&error);
    return (gdouble)(vali-mini)/(gdouble)(maxi-mini)*KEYSCALE;
  case WEED_HINT_FLOAT:
    vald=weed_get_double_value(in_param,"value",&error);
    mind=weed_get_double_value(paramtmpl,"min",&error);
    maxd=weed_get_double_value(paramtmpl,"max",&error);
    return (vald-mind)/(maxd-mind)*KEYSCALE;
  case WEED_HINT_SWITCH:
    vali=weed_get_boolean_value(in_param,"value",&error);
    return vali;
  }

  return 0;
}




















////////////////////////////////////////////////////////////////////////
gchar *weed_category_to_text(int cat) {
  // return value should be free'd after use
  switch (cat) {
  case 1:
    return (g_strdup(_("generator")));
  case 2:
    return (g_strdup(_("transition")));
  case 3:
    return (g_strdup(_("effect")));
  case 4:
    return (g_strdup(_("utility")));
  case 5:
    return (g_strdup(_("compositor")));
  case 6:
    return (g_strdup(_("tap")));
  case 7:
    return (g_strdup(_("splitter")));
  default:
    return (g_strdup(_("unknown")));
  }
}


gchar *weed_instance_get_type(gint idx) {
  // return value should be free'd after use
  return weed_category_to_text(weed_filter_categorise(enabled_in_channels(idx,FALSE),enabled_out_channels(idx,FALSE)));
}


gchar *weed_filter_get_type(gint idx) {
  // return value should be free'd after use
  return weed_category_to_text(weed_filter_categorise(enabled_in_channels(idx,TRUE),enabled_out_channels(idx,TRUE)));
}


gchar *rte_keymode_get_type (gint key, gint mode) {
  // return value should be free'd after use
  gchar *type=g_strdup ("");
  weed_plant_t *filter;
  gint idx,inst_idx;

  key--;

  if (key<0||key>FX_KEYS_PHYSICAL||mode<0||mode>=MAX_MODES_PER_KEY) return type;

  if ((idx=key_to_fx[key][mode])==-1) return type;
  if ((filter=weed_filters[idx])==NULL) return type;

  mainw->osc_block=TRUE;

  if ((inst_idx=key_to_instance[key][mode])!=-1) {
    // return details for instance
    g_free(type);
    type=weed_instance_get_type(inst_idx);
    mainw->osc_block=FALSE;
  }
  else type=weed_filter_get_type(idx);
  return type;
}

///////////////////////////////////////////////////////////////////////////////

int get_next_free_key(void) {
  // 0 based
  int i,free_key;
  free_key=next_free_key;
  for (i=free_key+1;i<FX_KEYS_MAX;i++) {
    if (key_to_fx[i][0]==-1) {
      next_free_key=i;
      break;
    }
  }
  if (i==FX_KEYS_MAX) next_free_key=-1;
  return free_key;
}


gboolean weed_delete_effectkey (gint key, gint mode) {
  // delete the effect binding for key/mode and move higher numbered slots down
  // also moves the active mode if applicable
  // returns FALSE if there was no effect bound to key/mode

  int oldkeymode=key_modes[--key];
  gint orig_mode=mode;
  gboolean was_started=FALSE;
  gint modekey=key;

  if (key_to_fx[key][mode]==-1) return FALSE;

  for (;mode<MAX_MODES_PER_KEY;mode++) {

    mainw->osc_block=TRUE;
    if (mode==MAX_MODES_PER_KEY-1||key_to_fx[key][mode+1]==-1) {
      if (key_to_instance[key][mode]!=-1) {
	was_started=TRUE;
	if (key_modes[key]==mode) modekey=-key-1;
	else key_modes[key]=mode;
	weed_deinit_effect(modekey);
	key_modes[key]=oldkeymode;
      }

      key_to_fx[key][mode]=-1;

      if (mode==orig_mode&&key_modes[key]==mode) {
	key_modes[key]=0;
	if (was_started) {
	  if (key_to_fx[key][0]!=-1) weed_init_effect(modekey);
	  else if (mainw->rte&(GU641<<key)) mainw->rte^=(GU641<<key);
	}
      }
      break; // quit the loop
    }
    else {
      rte_switch_keymode(key+1,mode,make_weed_hashname(key_to_fx[key+1][mode+1]));
    }
  }

  if (key>=FX_KEYS_MAX_VIRTUAL&&key<next_free_key) next_free_key=key;

  mainw->osc_block=FALSE;
  if (key_modes[key]>orig_mode) key_modes[key]--;

  return TRUE;
}








/////////////////////////////////////////////////////////////////////////////

gboolean rte_key_valid (int hotkey) {
  if (hotkey<0||hotkey>FX_KEYS_MAX) return FALSE;
  if (key_to_fx[hotkey][key_modes[hotkey]]==-1) return FALSE;
  return TRUE;
}

gboolean rte_keymode_valid (gint key, gint mode) {
  if (key_to_fx[--key][mode]==-1) return FALSE;
  return TRUE;
}

gint rte_keymode_get_filter_idx(gint key, gint mode) {
  return (key_to_fx[--key][mode]);
}

int rte_key_getmode (gint key) {
  return key_modes[--key];
}

weed_plant_t *rte_keymode_get_instance(gint key, gint mode) {
  weed_plant_t *inst;

  key--;
  if (!rte_keymode_valid(key+1,mode)) return NULL;
  mainw->osc_block=TRUE;
  if (key_to_instance[key][mode]==-1) {
    mainw->osc_block=FALSE;
    return NULL;
  }
  inst=weed_instances[key_to_instance[key][mode]];
  mainw->osc_block=FALSE;
  return inst;
}


weed_plant_t *rte_keymode_get_filter(gint key, gint mode) {
  key--;
  if (!rte_keymode_valid(key+1,mode)) return NULL;
  return weed_filters[key_to_fx[key][mode]];
}


gchar *weed_filter_get_name(gint idx) {
  // return value should be free'd after use
  weed_plant_t *filter;
  int error;
  gchar *filter_name,*retval;

  if (idx==-1) return g_strdup("");
  if ((filter=weed_filters[idx])==NULL) return g_strdup("");
  filter_name=weed_get_string_value(filter,"name",&error);
  retval=g_strdup(filter_name);
  weed_free(filter_name);
  return retval;
}


gchar *rte_keymode_get_filter_name (gint key, gint mode) {
  // return value should be free'd after use
  gchar *filter_name,*retval;
  key--;

  if (key<0||key>FX_KEYS_PHYSICAL||mode<0||mode>=MAX_MODES_PER_KEY) return g_strdup("");
  filter_name=weed_filter_get_name(key_to_fx[key][mode]);
  retval=g_strdup(filter_name);
  return retval;
}


gchar *rte_keymode_get_plugin_name(gint key, gint mode) {
  // return value should be g_free'd after use
  weed_plant_t *filter,*plugin_info;
  gchar *name;
  int error;

  key--;
  if (!rte_keymode_valid(key+1,mode)) return g_strdup("");

  filter=weed_filters[key_to_fx[key][mode]];
  plugin_info=weed_get_plantptr_value(filter,"plugin_info",&error);
  name=weed_get_string_value(plugin_info,"name",&error);
  return g_strdup(name);
}




int rte_getmodespk (void) {
  return MAX_MODES_PER_KEY;
}

gint rte_bg_gen_key (void) {
  return bg_generator_key;
}

gint rte_fg_gen_key (void) {
  return fg_generator_key;
}

gint rte_bg_gen_mode (void) {
  return bg_generator_mode;
}

gint rte_fg_gen_mode (void) {
  return fg_generator_mode;
}


gboolean rte_key_setmode (gint key, gint newmode) {
  gint oldmode;
  gint blend_file;
  gshort whentostop=mainw->whentostop;
  gboolean was_started=FALSE;
  gint idx,real_key;

  if (key==0) {
    if ((key=mainw->rte_keys)==-1) return FALSE;
  }
  else key--;

  real_key=key;

  oldmode=key_modes[key];

  if (newmode==-1) {
    if (oldmode==MAX_MODES_PER_KEY||key_to_fx[key][oldmode+1]==-1) {
      newmode=0;
    }
    else {
      newmode=key_modes[key]+1;
    }
  }

  if (newmode<0||newmode>=MAX_MODES_PER_KEY) return FALSE;

  if (key_to_fx[key][newmode]==-1) return FALSE;

  if (rte_window!=NULL) rtew_set_mode_radio(key,newmode);

  mainw->osc_block=TRUE;

  // TODO - block template channel changes

  if ((idx=key_to_instance[key][oldmode])!=-1&&weed_instances[idx]!=NULL){
    was_started=TRUE;
    if (enabled_in_channels(idx,FALSE)==2&&enabled_in_channels(key_to_fx[key][newmode],TRUE)==2) {
      // transition --> transition, allow any bg generators to survive
      key=-key-1;
    }
  }

  if (oldmode!=newmode) {
    blend_file=mainw->blend_file;

    if (was_started&&enabled_in_channels(idx,FALSE)>0) {
      // not a generator
      weed_deinit_effect (key);
    }
    else if (enabled_in_channels(key_to_fx[key][newmode],TRUE)==0) mainw->whentostop=NEVER_STOP; // when gen->gen, dont stop pb
    
    key_modes[real_key]=newmode;

    mainw->blend_file=blend_file;

    if (was_started) {
      if (!weed_init_effect(key)) {
	// TODO - unblock template channel changes
	mainw->whentostop=whentostop;
	key=real_key;
	if (mainw->rte&(GU641<<key)) mainw->rte^=(GU641<<key);
	mainw->osc_block=FALSE;
	return FALSE;
       }
    }
    // TODO - unblock template channel changes
    mainw->whentostop=whentostop;
  }

  mainw->osc_block=FALSE;
  return TRUE;
}


int weed_add_effectkey_by_idx (gint hotkey, int idx) {
  // we will add a filter_class at the next free slot for key, and return the slot number
  // if idx is -1 (probably meaning the filter was not found), we return -1
  // if all slots are full, we return -3
  // currently, generators and non-generators cannot be mixed on the same key (causes problems if the mode is switched)
  // in this case a -2 is returned
  gboolean has_gen=FALSE;
  gboolean has_non_gen=FALSE;

  int i;

  if (idx==-1) return -1;

  for (i=0;i<MAX_MODES_PER_KEY;i++) {
    if (key_to_fx[hotkey][i]!=-1) {
      if (enabled_in_channels(key_to_fx[hotkey][i],TRUE)==0) has_gen=TRUE;
      else has_non_gen=TRUE;
    }
    else {
      if ((enabled_in_channels(idx,TRUE)==0&&has_non_gen)||(enabled_in_channels(idx,TRUE)>0&&has_gen)) return -2;
      key_to_fx[hotkey][i]=idx;
      return i;
    }
  }
  return -3;
}


int weed_add_effectkey (gint key, const gchar *hashname) {
  // add a filter_class by hashname to an effect_key
  int idx=weed_get_idx_for_hashname(hashname);
  return weed_add_effectkey_by_idx (--key,idx);
}



gint rte_switch_keymode (gint key, gint mode, const gchar *hashname) {
  // this is called when we switch the filter_class bound to an effect_key/mode

  int oldkeymode=key_modes[--key];
  int id=weed_get_idx_for_hashname (hashname);
  gboolean osc_block;
  gboolean has_gen=FALSE,has_non_gen=FALSE;

  // effect not found
  if (id==-1) return -1;

  if (key_to_fx[key][0]!=-1) {
    if (enabled_in_channels(key_to_fx[key][0],TRUE)==0) has_gen=TRUE;
    else has_non_gen=TRUE;
  }

  if ((enabled_in_channels(id,TRUE)==0&&has_non_gen)||(enabled_in_channels(id,TRUE)>0&&has_gen)) return -2;

  osc_block=mainw->osc_block;
  mainw->osc_block=TRUE;
  if (key_to_instance[key][mode]!=-1) {
    key_modes[key]=mode;
    weed_deinit_effect(-key-1); // set is_modeswitch
    key_to_fx[key][mode]=id;
    weed_init_effect(-key-1);
    key_modes[key]=oldkeymode;
  }
  else key_to_fx[key][mode]=id;

  mainw->osc_block=osc_block;
  return 0;
}



void rte_swap_fg_bg (void) {
  gint key=fg_generator_key;
  gint mode=fg_generator_mode;

  if (key!=-1) {
    fg_generator_clip=-1;
  }
  fg_generator_key=bg_generator_key;
  fg_generator_mode=bg_generator_mode;
  if (fg_generator_key!=-1) {
    fg_generator_clip=mainw->current_file;
  }
  bg_generator_key=key;
  bg_generator_mode=mode;
}



GList *weed_get_all_names (gshort list_type) {
  // remember to free list after use, if non-NULL
  GList *list=NULL;
  int i,error;
  gchar *filter_name,*filter_type,*hashname,*string;

  for (i=0;i<num_weed_filters;i++) {
    filter_name=weed_get_string_value(weed_filters[i],"name",&error);
    switch (list_type) {
    case 1:
      // just name
      string=g_strdup(filter_name);
      list=g_list_append(list,(gpointer)string);
      break;
    case 2:
      // name and type
      filter_type=weed_filter_get_type(i);
      string=g_strdup_printf("%s (%s)",filter_name,filter_type);
      list=g_list_append(list,(gpointer)string);
      g_free(filter_type);
      break;
    case 3:
      // hashnames
      hashname=make_weed_hashname(i);
      list=g_list_append(list,(gpointer)hashname);
      break;
    }
    weed_free(filter_name);
  }
  return list;
}


gint rte_get_numeffects(void) {
  return num_weed_filters;
}







///////////////////////////////////////////////////////////
////// hashnames

gchar *make_weed_hashname(int filter_idx) {
  weed_plant_t *filter,*plugin_info;
  gchar *plugin_name,*filter_name,*filter_author,*filter_version,*hashname;
  int error,version;
  gchar plugin_fname[256];

  if (filter_idx<0||filter_idx>=num_weed_filters) return g_strdup("");

  if (hashnames[filter_idx]!=NULL) return g_strdup(hashnames[filter_idx]);

  filter=weed_filters[filter_idx];

  plugin_info=weed_get_plantptr_value(filter,"plugin_info",&error);

  plugin_name=weed_get_string_value(plugin_info,"name",&error);

  g_snprintf(plugin_fname,256,"%s",plugin_name);
  weed_free(plugin_name);
  get_filename(plugin_fname);

  filter_name=weed_get_string_value(filter,"name",&error);
  filter_author=weed_get_string_value(filter,"author",&error);
  version=weed_get_int_value(filter,"version",&error);
  filter_version=g_strdup_printf("%d",version);

  hashname=g_strconcat("Weed",plugin_fname,filter_name,filter_author,filter_version,NULL);

  weed_free(filter_name);
  weed_free(filter_author);
  g_free(filter_version);

  return hashname;
}



int weed_get_idx_for_hashname (const gchar *hashname) {
  int i;
  for (i=0;i<num_weed_filters;i++) {
    if (!strcmp(hashname,make_weed_hashname(i))) return i;
  }
  return -1;
}
