// events.c
// LiVES
// (c) G. Finch 2005 <salsaman@xs4all.nl>
// released under the GNU GPL 2
// see file ../COPYING or www.gnu.org for licensing details


// functions/structs for event_lists and events


#include "main.h"
#include "effects.h"
#include "support.h"
#include "callbacks.h"
#include "resample.h"
#include "paramwindow.h" // for DEF_BUTTON_WIDTH
#include "interface.h" // DEF frame sizes - TODO

//////////////////////////////////
//#define DEBUG_EVENTS


static gint render_choice;

///////////////////////////////////////////////////////

inline weed_timecode_t get_event_timecode (weed_plant_t *plant) {
  weed_timecode_t tc;
  int error;

  tc=weed_get_int64_value (plant,"timecode",&error);
  return tc;
}

inline int get_event_hint (weed_plant_t *plant) {
  int hint;
  int error;
  hint=weed_get_int_value (plant,"hint",&error);
  return hint;
}


int get_frame_event_clip (weed_plant_t *plant, int layer) {
  int numframes,clipnum;
  int *clips,error;
  if (get_event_hint (plant)!=LIVES_EVENT_HINT_FRAME) return -2;
  numframes=weed_leaf_num_elements (plant,"clips");
  clips=weed_get_int_array(plant,"clips",&error);
  if (numframes<=layer) return -3;
  clipnum=clips[layer];
  weed_free(clips);
  return clipnum;
}

static void delete_event (event_list_t *event_list, event_t *event) {
  if (event->prev!=NULL) event->prev->next=event->next;
  if (event->next!=NULL) event->next->prev=event->prev;

  if (event==event_list->first) event_list->first=event->next;
  if (event==event_list->last) event_list->last=event->prev;

  weed_plant_free (event->event);
  g_free (event);
}


event_t *get_next_frame_event (event_t *event) {
  event_t *next_event;
  if (event==NULL) return NULL;
  next_event=event->next;
  while (next_event!=NULL) {
    if (get_event_hint (next_event->event)==LIVES_EVENT_HINT_FRAME) return next_event;
    next_event=next_event->next;
  }
  return NULL;
}


event_t *get_prev_frame_event (event_t *event) {
  event_t *prev_event;
  if (event==NULL) return NULL;
  prev_event=event->prev;
  while (prev_event!=NULL) {
    if (get_event_hint (prev_event->event)==LIVES_EVENT_HINT_FRAME) return prev_event;
    prev_event=prev_event->prev;
  }
  return NULL;
}


event_t *get_first_frame_event (event_list_t *event_list) {
  event_t *event;

  if (event_list==NULL) return NULL;

  event=event_list->first;

  while (event!=NULL) {
    if (get_event_hint(event->event)==LIVES_EVENT_HINT_FRAME) return event;
    event=event->next;
  }
  return NULL;
}

event_t *get_last_frame_event (event_list_t *event_list) {
  event_t *event;

  if (event_list==NULL) return NULL;

  event=event_list->last;

  while (event!=NULL) {
    if (get_event_hint(event->event)==LIVES_EVENT_HINT_FRAME) return event;
    event=event->prev;
  }
  return NULL;
}


void *find_init_event_by_id(void *init_event, event_t *event) {
  void *event_id;
  int error;

  while (event!=NULL) {
    if (get_event_hint(event->event)==LIVES_EVENT_HINT_EFFECT_INIT) {
      event_id=weed_get_voidptr_value (event->event,"event_id",&error);
      if (event_id==init_event) break;
    }
    event=event->prev;
  }
  return (void *)event;
}


void remove_frame_from_event (event_list_t *event_list, event_t *event, gint track) {
  // TODO - memcheck
  int *clips;
  int *frames;
  weed_timecode_t tc;
  int numframes;
  int i,error;

  if (get_event_hint(event->event)!=LIVES_EVENT_HINT_FRAME) return;

  tc=get_event_timecode(event->event);

  numframes=weed_leaf_num_elements(event->event,"clips");
  clips=weed_get_int_array(event->event,"clips",&error);
  frames=weed_get_int_array(event->event,"frames",&error);

  if (track==numframes-1) numframes--;
  else {
    clips[track]=-1;
    frames[track]=0;
  }

  // if stack is empty, we will replace with a blank frame
  for (i=0;i<numframes&&clips[i]<1;i++);
  if (i==numframes) {
    int blank_clip=-1,blank_frame=0;
    if (event==event_list->last) delete_event(event_list,event);
    else insert_frame_event_at (event_list,tc,1,&blank_clip,&blank_frame,&event);
  }
  else insert_frame_event_at (event_list,tc,numframes,clips,frames,&event);
  weed_free(frames);
  weed_free(clips);
}



gboolean is_blank_frame (weed_plant_t *event) {
  int clip,frame,numframes,error;
  if (get_event_hint(event)!=LIVES_EVENT_HINT_FRAME) return FALSE;
  numframes=weed_leaf_num_elements(event,"clips");
  if (numframes>1) return FALSE;
  clip=weed_get_int_value(event,"clips",&error);
  frame=weed_get_int_value(event,"frames",&error);

  if (clip==-1&&frame==0) return TRUE;
  return FALSE;
}


void remove_end_blank_frames (event_list_t *event_list) {
  // remove blank frames from end of event list
  event_t *event=event_list->last,*prevevent;
  while (event!=NULL) {
    if (!is_blank_frame(event->event)) break;
    prevevent=event->prev;
    delete_event(event_list,event);
    event=prevevent;
  }
}


event_list_t *event_copy_and_append (event_t *in_eventx, event_list_t *event_list) {
  // this is called during quantisation

  // copy a non-frame event and append it to event_list
  // events must be copied in time order, since effect_deinit, effect_map and param_change events refer to prior effect_init events

  // when we copy an effect_init, we add a new field to the copy "event_id". This contains the value of the pointer to original event
  // we use this to locate the effect_init event in the new event_list

  // for effect_deinit, effect_map, param_change, we change the "init_event(s)" property to point to our copy effect_init

  weed_plant_t *event,*in_event;
  gboolean is_first=FALSE;
  event_t *lives_event;
  weed_timecode_t tc;
  int hint;
  int *tracks;
  int num_tracks,num_events,i;
  void *init_event,**init_events;
  char *string;
  int error;

  if (in_eventx==NULL) return event_list;
  in_event=in_eventx->event;
  if (in_event==NULL) return event_list;

  if (event_list==NULL) {
    if ((event_list=g_malloc (sizeof(event_list_t)))==NULL) return NULL;
    event_list->first=event_list->last=NULL;
  }

  if (event_list->first==NULL) is_first=TRUE;

  event=weed_plant_new (WEED_PLANT_EVENT);
  if (event==NULL) return NULL;

  tc=get_event_timecode(in_event);
  hint=get_event_hint(in_event);

  tc=(gint)(tc/U_SEC*cfile->fps+.5)/cfile->fps*U_SECL;

  weed_set_int64_value (event,"timecode",tc);
  weed_set_int_value (event,"hint",hint);

  switch (hint) {
  case LIVES_EVENT_HINT_EFFECT_INIT :
    // TODO - error check
    string=weed_get_string_value(in_event,"filter",&error);
    weed_set_string_value (event,"filter",string);
    weed_free(string);

    if (weed_plant_has_leaf(in_event,"host_tag")) {
      string=weed_get_string_value(in_event,"host_tag",&error);
      weed_set_string_value (event,"host_tag",string);
      g_free(string);
    }

    if (weed_plant_has_leaf(in_event,"in_tracks")) {
      num_tracks=weed_leaf_num_elements (in_event,"in_tracks");
      tracks=weed_get_int_array (in_event,"in_tracks",&error);
      weed_set_int_array (event,"in_tracks",num_tracks,tracks);
      weed_free(tracks);
    }

    if (weed_plant_has_leaf(in_event,"out_tracks")) {
      num_tracks=weed_leaf_num_elements (in_event,"out_tracks");
      tracks=weed_get_int_array (in_event,"out_tracks",&error);
      weed_set_int_array (event,"out_tracks",num_tracks,tracks);
      weed_free(tracks);
    }

    weed_set_voidptr_value (event,"event_id",(void *)in_eventx);
    break;
  case LIVES_EVENT_HINT_EFFECT_DEINIT:
    init_event=weed_get_voidptr_value (in_event,"init_event",&error);
    weed_set_voidptr_value (event,"init_event",find_init_event_by_id(init_event,event_list->last));
    break;
  case LIVES_EVENT_HINT_EFFECT_MAP:
    // set "init_events" property
    num_events=weed_leaf_num_elements (in_event,"init_events");
    init_events=weed_get_voidptr_array (in_event,"init_events",&error);
    for (i=0;i<num_events;i++) {
      init_events[i]=find_init_event_by_id(init_events[i],event_list->last);
    }
    weed_set_voidptr_array (event,"init_events",num_events,init_events);
    weed_free(init_events);
    break;
  case LIVES_EVENT_HINT_PARAM_CHANGE:

    break;
  }

  lives_event=g_malloc (sizeof(event_t));
  lives_event->next=NULL;
  lives_event->event=event;

  if (is_first) {
    event_list->first=lives_event;
    lives_event->prev=NULL;
  }
  else {
    lives_event->prev=event_list->last;
  }
  if (lives_event->prev!=NULL) lives_event->prev->next=lives_event;
  event_list->last=lives_event;

  return event_list;
}


event_t *get_frame_event_at (event_list_t *event_list, weed_timecode_t tc, event_t *shortcut) {
  event_t *event;
  weed_timecode_t xtc;

  if (event_list==NULL) return NULL;
  if (shortcut!=NULL) event=shortcut;
  else event=event_list->first;
  while (event!=NULL) {
    if (tc==(xtc=get_event_timecode(event->event))&&get_event_hint(event->event)==LIVES_EVENT_HINT_FRAME) return event;
    else if (xtc>tc) return NULL;
    event=event->next;
  }
  return NULL;
}





event_list_t *insert_frame_event_at (event_list_t *event_list, weed_timecode_t tc, int numframes, int *clips, int *frames, event_t **shortcut) {
  // we will insert a FRAME event at timecode tc, after any other events at timecode tc
  // if there is an existing FRAME event at tc, we replace it with the new frame

  // shortcut can be a nearest guess of where the frame should be

  event_t *event=NULL,*new_event;
  event_list_t *new_event_list;
  weed_timecode_t xtc;

  if (event_list==NULL) {
    event_list=append_frame_event(event_list,tc,numframes,clips,frames);
    if (shortcut!=NULL) *shortcut=event_list->last;
    return event_list;
  }

  if (shortcut!=NULL&&*shortcut!=NULL) {
    event=*shortcut;
    while (event!=NULL&&get_event_timecode(event->event)>tc) event=get_prev_frame_event(event);
  }
  if (event==NULL) event=event_list->first;

  while (event!=NULL&&((xtc=get_event_timecode(event->event))<=tc)) {
    if (shortcut!=NULL) *shortcut=event;
    if (xtc==tc&&get_event_hint(event->event)==LIVES_EVENT_HINT_FRAME) {
      // TODO - memcheck
      weed_set_int_array (event->event,"clips",numframes,clips);
      weed_set_int_array (event->event,"frames",numframes,frames);
      return event_list;
    }
    event=event->next;
  }
  if (event==NULL) {
    append_frame_event (event_list,tc,numframes,clips,frames);
    if (shortcut!=NULL) *shortcut=event_list->last;
    return event_list;
  }

  // OK, event is the event after where we want to insert our frame event
  
  if ((new_event_list=append_frame_event(NULL,tc,numframes,clips,frames))==NULL) return NULL;
  // new_event_list is now an event_list with one frame event. We will steal its event and prepend it !

  new_event=new_event_list->first;
  if (event->prev!=NULL) event->prev->next=new_event;
  new_event->prev=event->prev;
  new_event->next=event;
  event->prev=new_event;

  if (event_list->first==event) event_list->first=new_event;

  g_free(new_event_list);

  if (shortcut!=NULL) *shortcut=new_event;
  return event_list;

}


//////////////////////////////////////////////////////

void
set_render_choice (GtkToggleButton *togglebutton, gpointer choice) {
  if (gtk_toggle_button_get_active(togglebutton)) render_choice=GPOINTER_TO_INT (choice);
}

void
set_render_choice_button (GtkButton *button, gpointer choice) {
  render_choice=GPOINTER_TO_INT (choice);
}


gint get_render_choice (void) {
  return render_choice;
}




GtkWidget *events_rec_dialog (void) {
  GtkWidget *e_rec_dialog;
  GtkWidget *dialog_vbox;
  GtkWidget *dialog_action_area;
  GtkWidget *vbox;
  GtkWidget *label;
  GtkWidget *radiobutton;
  GtkWidget *okbutton;
  GtkWidget *cancelbutton;
  GSList *radiobutton_group = NULL;
  GtkAccelGroup *accel_group;

  render_choice=RENDER_CHOICE_PREVIEW;

  e_rec_dialog = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (e_rec_dialog), _("LiVES: - Events recorded"));
  gtk_window_set_position (GTK_WINDOW (e_rec_dialog), GTK_WIN_POS_CENTER);
  gtk_window_set_modal (GTK_WINDOW (e_rec_dialog), TRUE);
  if (palette->style&STYLE_3||palette->style==STYLE_PLAIN) {
    gtk_widget_modify_bg (e_rec_dialog, GTK_STATE_NORMAL, &palette->normal_back);
  }
  else {
    gtk_widget_modify_bg (e_rec_dialog, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }
  gtk_window_set_transient_for(GTK_WINDOW(e_rec_dialog),GTK_WINDOW(mainw->LiVES));
  gtk_window_set_modal (GTK_WINDOW (e_rec_dialog), TRUE);

  dialog_vbox = GTK_DIALOG (e_rec_dialog)->vbox;
  gtk_widget_show (dialog_vbox);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_box_pack_start (GTK_BOX (dialog_vbox), vbox, TRUE, TRUE, 0);

  label = gtk_label_new (_("   Events were recorded. What would you like to do with them ?    "));
  gtk_box_pack_start (GTK_BOX (vbox), label, TRUE, TRUE, 0);
  gtk_widget_show (label);


  radiobutton = gtk_radio_button_new_with_mnemonic (NULL, _("_Preview events"));
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton), radiobutton_group);
  radiobutton_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton));
  gtk_container_set_border_width (GTK_CONTAINER (radiobutton), 18);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radiobutton),TRUE);
  gtk_box_pack_start (GTK_BOX (vbox), radiobutton, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS (radiobutton, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
  g_signal_connect (GTK_OBJECT (radiobutton), "toggled",
		    G_CALLBACK (set_render_choice),
		    GINT_TO_POINTER (RENDER_CHOICE_PREVIEW));

  if (!mainw->clip_switched) {
    radiobutton = gtk_radio_button_new_with_mnemonic (NULL, _("Render events to _same clip"));
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton), radiobutton_group);
    radiobutton_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton));
    gtk_container_set_border_width (GTK_CONTAINER (radiobutton), 18);
    gtk_box_pack_start (GTK_BOX (vbox), radiobutton, TRUE, TRUE, 0);
    GTK_WIDGET_SET_FLAGS (radiobutton, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
    g_signal_connect (GTK_OBJECT (radiobutton), "toggled",
                      G_CALLBACK (set_render_choice),
                      GINT_TO_POINTER (RENDER_CHOICE_SAME_CLIP));
  }

  radiobutton = gtk_radio_button_new_with_mnemonic (NULL, _("Render events to _new clip"));
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton), radiobutton_group);
  radiobutton_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton));
  gtk_container_set_border_width (GTK_CONTAINER (radiobutton), 18);
  gtk_box_pack_start (GTK_BOX (vbox), radiobutton, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS (radiobutton, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
  g_signal_connect (GTK_OBJECT (radiobutton), "toggled",
		    G_CALLBACK (set_render_choice),
		    GINT_TO_POINTER (RENDER_CHOICE_NEW_CLIP));
  
  radiobutton = gtk_radio_button_new_with_mnemonic (NULL, _("View/edit events in _multitrack window (test)"));
#ifndef TEST_MT
  gtk_widget_set_sensitive (radiobutton, FALSE);
#endif
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton), radiobutton_group);
  radiobutton_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton));
  gtk_container_set_border_width (GTK_CONTAINER (radiobutton), 18);
  gtk_box_pack_start (GTK_BOX (vbox), radiobutton, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS (radiobutton, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
  g_signal_connect (GTK_OBJECT (radiobutton), "toggled",
		    G_CALLBACK (set_render_choice),
		    GINT_TO_POINTER (RENDER_CHOICE_MULTITRACK));

  radiobutton = gtk_radio_button_new_with_mnemonic (NULL, _("View/edit events in _event window"));
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton), radiobutton_group);
  radiobutton_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton));
  gtk_container_set_border_width (GTK_CONTAINER (radiobutton), 18);
  gtk_box_pack_start (GTK_BOX (vbox), radiobutton, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS (radiobutton, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
  g_signal_connect (GTK_OBJECT (radiobutton), "toggled",
		    G_CALLBACK (set_render_choice),
		    GINT_TO_POINTER (RENDER_CHOICE_EVENT_LIST));

  dialog_action_area = GTK_DIALOG (e_rec_dialog)->action_area;
  cancelbutton = gtk_button_new_from_stock ("gtk-cancel");
  GTK_WIDGET_SET_FLAGS (cancelbutton, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
  gtk_dialog_add_action_widget (GTK_DIALOG (e_rec_dialog), cancelbutton, GTK_RESPONSE_CANCEL);

  g_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
		    G_CALLBACK (set_render_choice_button),
		    GINT_TO_POINTER (RENDER_CHOICE_DISCARD));

  accel_group = GTK_ACCEL_GROUP(gtk_accel_group_new ());
  gtk_widget_add_accelerator (cancelbutton, "activate", accel_group,
                              GDK_Escape, 0, 0);
  gtk_window_add_accel_group (GTK_WINDOW (e_rec_dialog), accel_group);

  okbutton = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (okbutton);
  gtk_dialog_add_action_widget (GTK_DIALOG (e_rec_dialog), okbutton, GTK_RESPONSE_OK);
  GTK_WIDGET_SET_FLAGS (okbutton, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
  gtk_widget_grab_default (okbutton);
  gtk_widget_show_all (e_rec_dialog);

  return e_rec_dialog;

}



static void event_list_free_events (event_list_t *event_list) {
  event_t *event,*next_event;
  event=event_list->first;

  while (event!=NULL) {
    weed_plant_free (event->event);
    next_event=event->next;
    g_free (event);
    event=next_event;
  }
}

void event_list_free (event_list_t *event_list) {
  if (event_list==NULL) return;
  event_list_free_events(event_list);
  g_free (event_list);
}

void event_list_replace_events (event_list_t *event_list, event_list_t *new_event_list) {
  if (event_list==NULL) return;
  event_list_free_events(event_list);
  event_list->first=new_event_list->first;
  event_list->last=new_event_list->last;
}




gint event_list_to_block (event_list_t *event_list) {
  // translate our new event_list to older event blocks
  // by now we should have eliminated clip switches and param settings

  // first we count the frame events
  int i=0,num_events,error;
  event_t *event;

  if (event_list==NULL) return 0;

  num_events=count_events (event_list,FALSE);

  // then we create event_frames

  if (!create_event_space(num_events)) {
    do_memory_error_dialog();
    return 0;
  }

  event=event_list->first;

  while (event!=NULL) {
    if (get_event_hint(event->event)==LIVES_EVENT_HINT_FRAME) {
      (cfile->events[0]+i++)->value=weed_get_int_value (event->event,"frames",&error);
    }
    event=event->next;
  }
  return num_events;
}



void event_list_add_offset (weed_timecode_t ticks, event_list_t *event_list) {
  // add a constant timecode offset to an event_list
  event_t *event;
  weed_timecode_t tc;

  if (event_list==NULL) return;
  
  event=event_list->first;
  while (event!=NULL) {
    tc=get_event_timecode (event->event);
    tc+=ticks;
    weed_set_int64_value (event->event,"timecode",tc);
    event=event->next;
  }
}


void event_list_add_track (event_list_t *event_list, gint layer) {
  // in this function we insert a new track before existing tracks
  // TODO - memcheck
  event_t *event;
  gint numframes;
  int *clips,*frames,*newclips,*newframes,i,error;

  if (event_list==NULL) return;

  event=event_list->first;
  while (event!=NULL) {
    switch (get_event_hint(event->event)) {
    case LIVES_EVENT_HINT_FRAME:
      numframes=weed_leaf_num_elements(event->event,"clips");
      clips=weed_get_int_array(event->event,"clips",&error);
      frames=weed_get_int_array(event->event,"frames",&error);
      if (numframes==1&&clips[0]==-1&&frames[0]==0) {
	// for blank frames, we don't do anything
	weed_free(clips);
	weed_free(frames);
	break;
      }

      newclips=(int *)g_malloc((numframes+1)*sizint);
      newframes=(int *)g_malloc((numframes+1)*sizint);
      
      newclips[layer]=-1;
      newframes[layer]=0;
      for (i=0;i<numframes;i++) {
	if (i<layer) {
	  newclips[i]=clips[i];
	  newframes[i]=frames[i];
	}
	else {
	  newclips[i+1]=clips[i];
	  newframes[i+1]=frames[i];
	}
      }
      numframes++;

      weed_set_int_array(event->event,"clips",numframes,newclips);
      weed_set_int_array(event->event,"frames",numframes,newframes);
      
      g_free(newclips);
      g_free(newframes);
      weed_free(clips);
      weed_free(frames);
      break;
    case LIVES_EVENT_HINT_EFFECT_INIT:
      break;
    }
    event=event->next;
  }
}


event_list_t *append_frame_event (event_list_t *event_list, weed_timecode_t tc, int numframes, int *clips, int *frames) {
  // append a frame event to an event_list

  weed_plant_t *event;
  event_t *lives_event;

  ///////////// to func //////////////
  if (event_list==NULL) {
    if ((event_list=g_malloc (sizeof(event_list_t)))==NULL) return NULL;
    event_list->first=event_list->last=NULL;
  }

  event=weed_plant_new (WEED_PLANT_EVENT);
  if (event==NULL) return NULL;
  ////////////////////////////////////////

  // TODO - error check
  weed_set_int64_value (event,"timecode",tc);
  weed_set_int_value (event,"hint",LIVES_EVENT_HINT_FRAME);

  weed_set_int_array (event,"clips",numframes,clips);
  weed_set_int_array (event,"frames",numframes,frames);

  ////////////////////// to func //////////
  lives_event=g_malloc (sizeof(event_t));
  lives_event->next=NULL;
  lives_event->event=event;

  if (event_list->first==NULL) {
    event_list->first=lives_event;
    lives_event->prev=NULL;
  }
  else {
    lives_event->prev=event_list->last;
  }
  if (lives_event->prev!=NULL) lives_event->prev->next=lives_event;
  event_list->last=lives_event;
  //////////////////////////////////////

  return event_list;
}


event_list_t *append_effect_init_event (event_list_t *event_list, weed_timecode_t tc, int filter_idx) {
  weed_plant_t *event;
  event_t *lives_event;
  gint num_in_channels;

  if (event_list==NULL) {
    if ((event_list=g_malloc (sizeof(event_list_t)))==NULL) return NULL;
    event_list->first=event_list->last=NULL;
  }

  event=weed_plant_new (WEED_PLANT_EVENT);
  if (event==NULL) return NULL;

  // TODO - error check
  weed_set_int64_value (event,"timecode",tc);
  weed_set_int_value (event,"hint",LIVES_EVENT_HINT_EFFECT_INIT);
  weed_set_string_value (event,"filter",make_weed_hashname (filter_idx));

  //////////////////////////////////////////////////////////////////////////////////

  num_in_channels=enabled_in_channels (filter_idx,TRUE);

  // ASSUME there are no unused channels - TODO **


  // here we map our tracks to channels
  if (num_in_channels==1) {
    weed_set_int_value (event,"in_tracks",0);
  }
  else {
    int *tracks=g_malloc (2*sizint);
    tracks[0]=0;
    tracks[1]=1;
    weed_set_int_array (event,"in_tracks",2,tracks);
    g_free (tracks);
  }

  weed_set_int_value (event,"out_tracks",0);


  ///////////////////////////////////////////////////////////////////////////////////
#ifdef DEBUG_EVENTS
  g_print ("adding init event at tc %ld\n",tc);
#endif

  lives_event=g_malloc (sizeof(event_t));
  lives_event->next=NULL;
  lives_event->event=event;

  if (event_list->first==NULL) {
    event_list->first=lives_event;
    lives_event->prev=NULL;
  }
  else {
    lives_event->prev=event_list->last;
  }
  if (lives_event->prev!=NULL) lives_event->prev->next=lives_event;
  event_list->last=lives_event;

  return event_list;

}


gboolean add_effect_deinit_event (event_list_t *event_list, weed_timecode_t tc, event_t *init_event) {
  weed_plant_t *event;
  event_t *lives_event;

  if (event_list==NULL) {
    if ((event_list=mainw->event_list=g_malloc (sizeof(event_list_t)))==NULL) return FALSE;
    mainw->event_list->first=mainw->event_list->last=NULL;
  }

  event=weed_plant_new (WEED_PLANT_EVENT);
  if (event==NULL) return FALSE;

  // TODO - error check
  weed_set_int64_value (event,"timecode",tc);
  weed_set_int_value (event,"hint",LIVES_EVENT_HINT_EFFECT_DEINIT);
  weed_set_voidptr_value (event,"init_event",(void *)init_event);

  lives_event=g_malloc (sizeof(event_t));
  lives_event->next=NULL;
  lives_event->event=event;

  if (event_list->first==NULL) {
    event_list->first=lives_event;
    lives_event->prev=NULL;
  }
  else {
    lives_event->prev=event_list->last;
  }
  if (lives_event->prev!=NULL) lives_event->prev->next=lives_event;
  event_list->last=lives_event;

  return TRUE;

}




gboolean add_effect_map_event (event_list_t *event_list, weed_timecode_t tc, event_t **init_events) {
  weed_plant_t *event;
  event_t *lives_event;
  int i;

  if (event_list==NULL) {
    if ((event_list=mainw->event_list=g_malloc (sizeof(event_list_t)))==NULL) return FALSE;
    mainw->event_list->first=mainw->event_list->last=NULL;
  }

  event=weed_plant_new (WEED_PLANT_EVENT);
  if (event==NULL) return FALSE;

  // TODO - error check
  weed_set_int64_value (event,"timecode",tc);
  weed_set_int_value (event,"hint",LIVES_EVENT_HINT_EFFECT_MAP);

  for (i=0;init_events[i]!=NULL;i++);

  weed_set_voidptr_array (event,"init_events",i,(void **)init_events);

#ifdef DEBUG_EVENTS
  g_print ("adding map event %p at tc %ld\n",init_events[0],tc);
#endif

  lives_event=g_malloc (sizeof(event_t));
  lives_event->next=NULL;
  lives_event->event=event;

  if (event_list->first==NULL) {
    event_list->first=lives_event;
    lives_event->prev=NULL;
  }
  else {
    lives_event->prev=event_list->last;
  }
  if (lives_event->prev!=NULL) lives_event->prev->next=lives_event;
  event_list->last=lives_event;

  return TRUE;
}





event_t *process_events (event_t *next_event, weed_timecode_t curr_tc, int *frameno) {
  // here we play back (preview) with an event_list
  // we process all events, but drop frames (unless mainw->nodrop is set)

  weed_timecode_t tc,next_tc;
  int hint;
  int key,idx;
  char *key_string;
  event_t *init_event,*next_frame_event,*return_event;
  int num_elems;
  int i;
  char *filter_name;
  int new_file;
  int num_tracks;
  gint current_file;
  int *clip_index,*frame_index;
  int error;

  if (next_event==NULL) return NULL;
  
  tc=get_event_timecode (next_event->event);
  if (tc>curr_tc) return next_event;

  return_event=next_event->next;

  hint=get_event_hint (next_event->event);
  switch (hint) {
  case LIVES_EVENT_HINT_FRAME:
#ifdef DEBUG_EVENTS
    g_print ("event: frame event at tc %lld curr_tc=%lld\n",tc,curr_tc);
#endif
    if ((next_frame_event=get_next_frame_event (next_event))!=NULL) {
      next_tc=get_event_timecode (next_frame_event->event);
      if (!mainw->fs&&prefs->show_framecount) {
	cfile->pb_fps=U_SEC/(next_tc-tc);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(mainw->spinbutton_pb_fps),cfile->pb_fps);
      }
      if (!mainw->noframedrop&&next_tc<=curr_tc) break;
    }

    num_tracks=weed_leaf_num_elements (next_event->event,"clips");

    clip_index=weed_get_int_array (next_event->event,"clips",&error);
    frame_index=weed_get_int_array (next_event->event,"frames",&error);

    // TODO - rip all this out, pull the frames in effects-weed.c
    if (num_tracks>1) {
      mainw->blend_file=clip_index[1];
      if (mainw->blend_file>-1) mainw->files[mainw->blend_file]->frameno=frame_index[1];
    }
    else mainw->blend_file=-1;
    
    new_file=-1;
    for (i=0;i<num_tracks&&new_file==-1;i++) new_file=clip_index[i];
    if (i==2) mainw->blend_file=-1;

#ifdef DEBUG_EVENTS
    g_print ("event: front frame is %d tc %lld curr_tc=%lld\n",frame_index[new_file],tc,curr_tc);
#endif

    // handle case where new_file==-1: we must somehow create a blank frame in load_frame_image
    if (new_file==-1) new_file=mainw->multitrack->render_file;

    if (new_file!=mainw->current_file) {
      mainw->files[new_file]->frameno=frame_index[i-1];
      weed_free(clip_index);
      weed_free(frame_index);
	
      // TODO *** !!!!
      if (mainw->files[new_file]->event_list!=NULL) {
	event_list_free (mainw->files[new_file]->event_list);
      }
      mainw->files[new_file]->event_list=cfile->event_list;

      cfile->event_list=NULL;

      if (new_file!=mainw->scrap_file) {
	// switch to a new file
	mainw->noswitch=FALSE;
	do_quick_switch (new_file);
	mainw->noswitch=TRUE;
	cfile->next_event=next_event->next;
	return_event=NULL;
      }
      else {
	mainw->files[new_file]->hsize=cfile->hsize; // TODO ??!?
	mainw->files[new_file]->vsize=cfile->vsize;
	current_file=mainw->current_file;
	mainw->current_file=new_file;
	cfile->next_event=next_event->next;
	load_frame_image (cfile->frameno,cfile->frameno);
	while (g_main_context_iteration (NULL,FALSE));
	mainw->files[current_file]->event_list=cfile->event_list;
	cfile->event_list=NULL;
	mainw->current_file=current_file;
      }
      break;
    }
    else {
      if (mainw->multitrack!=NULL&&new_file==mainw->multitrack->render_file) {
	cfile->frameno=0; // will force blank frame creation
      }
      else cfile->frameno=frame_index[i-1];
      weed_free(clip_index);
      weed_free(frame_index);
      load_frame_image (cfile->frameno,cfile->frameno);
      while (g_main_context_iteration(NULL,FALSE));
    }
    break;
  case LIVES_EVENT_HINT_EFFECT_INIT:
    // effect init
    //  bind the weed_fx to next free key/0
    key=get_next_free_key();
    filter_name=weed_get_string_value (next_event->event,"filter",&error);
    
    // for now, assume we can locate the hashname
    idx=weed_get_idx_for_hashname (filter_name);
    weed_free (filter_name);
    weed_add_effectkey_by_idx (key,idx);
    key_string=g_strdup_printf ("%d",key);
    weed_set_string_value (next_event->event,"host_tag",key_string);
    g_free (key_string);
    
    //  init the effect - TODO - set param values
#ifdef DEBUG_EVENTS
    g_print ("event: init effect on key %d at tc %lld curr_tc=%lld\n",key,tc,curr_tc);
#endif
    weed_init_effect (key);
    break;

  case LIVES_EVENT_HINT_EFFECT_DEINIT:
    init_event=weed_get_voidptr_value (next_event->event,"init_event",&error);
    key_string=weed_get_string_value (init_event->event,"host_tag",&error);
    key=atoi (key_string);
    weed_free (key_string);
    weed_deinit_effect (key);
    weed_delete_effectkey (key+1,0);
    break;

  case LIVES_EVENT_HINT_EFFECT_MAP:
    mainw->filter_map=next_event->event;
    num_elems=weed_leaf_num_elements(next_event->event,"init_events");
#ifdef DEBUG_EVENTS
    g_print ("got new effect map with %d keys\n",num_elems);
#endif
    break;
  case LIVES_EVENT_HINT_PARAM_CHANGE:
    break;
  }
  return return_event;
}




gint render_events (gboolean reset) {
  // this is called repeatedly when we are rendering effect changes and/or clip switches
  // if we have clip switches we will resize and build a new clip
  
  static gint progress;
  static event_t *event,*eventnext;
  gint blend_file=mainw->blend_file;
  static gint frame;
  int hint;
  gchar oname[256];
  GError *error=NULL;
  weed_timecode_t tc;
  int key,idx;
  gchar *key_string,*com;
  gchar *filter_name;
  event_t *init_event;
  event_t *next_frame_event;
  weed_timecode_t next_tc=0,next_out_tc;
  static gint out_frame;
  GdkPixbuf *pixbuf;
  int num_tracks;
  int i;
  int *clip_index,*frame_index;
  int weed_error;
  weed_plant_t **layers;

  if (reset) {
    progress=frame=1;
    event=cfile->next_event;
    out_frame=(gint)((gdouble)(get_event_timecode (event->event)/U_SECL)*cfile->fps+1.);
    if (cfile->frames<out_frame) out_frame=cfile->frames+1;
    cfile->undo_start=out_frame;
    clear_mainw_msg();
    mainw->filter_map=NULL;
    return 1;
  }

  if (mainw->effects_paused) return 1;

  if (event!=NULL) {
    eventnext=event->next;

    // sig_progress...
    g_snprintf (mainw->msg,256,"%d",progress++);
    hint=get_event_hint (event->event);
    switch (hint) {
    case LIVES_EVENT_HINT_FRAME:
      tc=get_event_timecode (event->event);
      num_tracks=weed_leaf_num_elements (event->event,"clips");
      clip_index=weed_get_int_array (event->event,"clips",&weed_error);
      frame_index=weed_get_int_array (event->event,"frames",&weed_error);

      layers=g_malloc((num_tracks+1)*sizeof(weed_plant_t *));
      for (i=0;i<num_tracks;i++) {
	layers[i]=weed_plant_new(WEED_PLANT_CHANNEL);
	weed_set_int_value(layers[i],"clip",clip_index[i]);
	weed_set_int_value(layers[i],"frame",frame_index[i]);
	weed_set_int_value(layers[i],"current_palette",WEED_PALETTE_BGR24);
	weed_set_voidptr_value(layers[i],"pixel_data",NULL);
      }
      layers[i]=NULL;

      weed_free(clip_index);
      weed_free(frame_index);
      
      pixbuf=weed_apply_effects(layers,mainw->filter_map,cfile->hsize,cfile->vsize,tc);

      for (i=0;layers[i]!=NULL;i++) weed_plant_free(layers[i]);

      g_free(layers);
      mainw->blend_file=blend_file;
      
      next_frame_event=get_next_frame_event (event);
      
      // get tc of next frame event
      if (next_frame_event!=NULL) next_tc=get_event_timecode (next_frame_event->event);
      
      while (cfile->fps>0.) {
	next_out_tc=(weed_timecode_t)((out_frame-1.)/cfile->fps*U_SEC); // calculate tc of next out frame
	
	if (next_frame_event!=NULL) {
	  if (next_tc<next_out_tc||next_tc-next_out_tc<next_out_tc-tc) break;
	}
	else if (next_out_tc>tc) break;
	
	cfile->undo_end=out_frame;
	if (out_frame>cfile->frames) cfile->frames=out_frame;
	if (out_frame>cfile->end) cfile->end=out_frame;
	if (cfile->start==0) cfile->start=1;
	if (cfile->old_frames>0) g_snprintf(oname,256,"%s/%s/%08d.mgk",prefs->tmpdir,cfile->handle,out_frame++);
	if (!strcmp (prefs->image_ext,"jpg")) {
	  if (cfile->old_frames==0) g_snprintf(oname,256,"%s/%s/%08d.jpg",prefs->tmpdir,cfile->handle,out_frame++);
	  gdk_pixbuf_save (pixbuf, oname, "jpeg", &error,"quality", "100", NULL);
	}
	else if (!strcmp (prefs->image_ext,"png")) {
	  if (cfile->old_frames==0) g_snprintf(oname,256,"%s/%s/%08d.png",prefs->tmpdir,cfile->handle,out_frame++);
	  gdk_pixbuf_save (pixbuf, oname, "png", &error, NULL);
	}
	else {
	  //gdk_pixbuf_save_to_callback(...);
	}
      }
      gdk_pixbuf_unref (pixbuf);
      break;
    case LIVES_EVENT_HINT_EFFECT_INIT:
      // effect init
      //  bind the weed_fx to next free key/0
      key=get_next_free_key();
      filter_name=weed_get_string_value (event->event,"filter",&weed_error);
    
      // for now, assume we can find hashname
      idx=weed_get_idx_for_hashname (filter_name);
      weed_free (filter_name);
      weed_add_effectkey_by_idx (key,idx);
      key_string=g_strdup_printf ("%d",key);
      weed_set_string_value (event->event,"host_tag",key_string);
      g_free (key_string);

      //  init the effect - TODO - set param values
      weed_init_effect (key);
      break;
    case LIVES_EVENT_HINT_EFFECT_DEINIT:
      init_event=weed_get_voidptr_value (event->event,"init_event",&weed_error);
      key_string=weed_get_string_value (init_event->event,"host_tag",&weed_error);
      key=atoi (key_string);
      weed_free (key_string);
      weed_deinit_effect (key);
      weed_delete_effectkey (key+1,0);
      break;
    case LIVES_EVENT_HINT_PARAM_CHANGE:
      break;
    case LIVES_EVENT_HINT_EFFECT_MAP:
#ifdef DEBUG_EVENTS
      g_print ("got new effect map\n");
#endif
      mainw->filter_map=event->event;
      break;
    }
    event=eventnext;
  }
  else {
    if (cfile->old_frames==0) cfile->undo_start=cfile->undo_end=0;
    com=g_strdup_printf ("smogrify mv_mgk %s %d %d",cfile->handle,cfile->undo_start,cfile->undo_end);
    system (com);
    g_free (com);
    mainw->is_rendering=mainw->internal_messaging=FALSE;
    mainw->filter_map=NULL;
  }
 return 1;
}




gboolean start_render_effect_events (event_list_t *event_list) {
  // this is called to begin rendering effect events from an event_list into cfile
  // it will do a reorder/resample/resize/effect apply all in one pass

  gdouble old_pb_fps=cfile->pb_fps;
  gint oundo_start=cfile->undo_start;
  gint oundo_end=cfile->undo_end;

  if (event_list==NULL) return TRUE; //oh, that was easy !

  mainw->is_rendering=mainw->internal_messaging=TRUE;
  cfile->next_event=event_list->first;

  mainw->progress_fn=&render_events;
  mainw->progress_fn (TRUE);

  cfile->progress_start=1;
  cfile->progress_end=count_events (event_list,TRUE);

  cfile->pb_fps=1000000.;

  cfile->redoable=cfile->undoable=FALSE;
  gtk_widget_set_sensitive (mainw->redo, FALSE);
  gtk_widget_set_sensitive (mainw->undo, FALSE);

  cfile->undo_action=UNDO_RENDER;

  // play back the file as fast as possible, each time calling render_events()
  if ((!do_progress_dialog(TRUE,TRUE,"Rendering")&&mainw->cancelled!=CANCEL_KEEP)||mainw->error) {
    mainw->cancel_type=CANCEL_KILL;
    mainw->cancelled=0;

    if (mainw->error) {
      do_error_dialog (mainw->msg);
      d_print (_ ("cancelled.\n"));
    }
    cfile->undo_start=oundo_start;
    cfile->undo_end=oundo_end;
    cfile->pb_fps=old_pb_fps;
    mainw->internal_messaging=FALSE;
    mainw->resizing=FALSE;
    cfile->next_event=NULL;
    return FALSE;
  }

  mainw->cancel_type=CANCEL_KILL;
  mainw->cancelled=0;
  cfile->changed=TRUE;
  gtk_widget_set_sensitive (mainw->undo, TRUE);
  cfile->undoable=TRUE;
  cfile->pb_fps=old_pb_fps;
  gtk_widget_set_sensitive (mainw->select_last, TRUE);
  set_undoable (_ ("rendering"),TRUE);
  cfile->next_event=NULL;
  return TRUE;
}



gint count_events (event_list_t *event_list, gboolean all_events) {
  event_t *event;
  gint i=0;

  if (event_list==NULL) return 0;
  event=event_list->first;

  while (event!=NULL) {
    if (all_events||get_event_hint(event->event)==LIVES_EVENT_HINT_FRAME) i++;
    event=event->next;
  }
  return i;
}


gdouble event_list_get_end_secs (event_list_t *event_list) {
  weed_timecode_t tc;
  if (event_list==NULL||event_list->last==NULL) return 0.;
  tc=get_event_timecode (event_list->last->event);
  return (tc/U_SEC);
}


gdouble event_list_get_start_secs (event_list_t *event_list) {
  weed_timecode_t tc;
  if (event_list==NULL||event_list->first==NULL) return 0.;
  tc=get_event_timecode (event_list->first->event);
  return (tc/U_SEC);
}






///////////////////////////////////////////////////////////////////


gboolean render_to_clip (gboolean new_clip) {
  // this function is called to actually start rendering mainw->event_list to a new/current clip
  gboolean retval=TRUE;
  GtkWidget *rdet_dialog;
  render_details rdet;
  gint current_file=mainw->current_file;
  gboolean response;

  d_print("Rendering...");

  if (new_clip) {
    //set file details
    rdet_dialog=create_render_details_dialog(&rdet);
    response=gtk_dialog_run(GTK_DIALOG(rdet_dialog));
    gtk_widget_destroy (rdet_dialog);

    if (response==GTK_RESPONSE_CANCEL) return FALSE;

    // create new file
    mainw->current_file=mainw->first_free_file;

    if (!get_new_handle(mainw->current_file,NULL)) {
      // bummer...
      mainw->current_file=current_file;
      return FALSE; // show dialog again
    }
    cfile->hsize=rdet.width;
    cfile->vsize=rdet.height;
    cfile->pb_fps=cfile->fps=rdet.fps;
    cfile->bpp=24;
    cfile->changed=TRUE;
    cfile->is_loaded=TRUE;
  }

  cfile->old_frames=cfile->frames;
  mainw->effects_paused=FALSE;
  if (start_render_effect_events(mainw->event_list)) { // re-render, applying effects 
    // and reordering/resampling/resizing if necessary
    //if (cfile->event_list_back!=NULL) event_list_free (cfile->event_list_back);
    //cfile->event_list_back=mainw->event_list;
    if (mainw->event_list!=NULL) event_list_free(mainw->event_list); // TODO
    mainw->event_list=NULL;
    if (new_clip) {
      cfile->start=1;
      cfile->end=cfile->frames;
      set_undoable (NULL,FALSE);
      add_to_winmenu();
      current_file=mainw->current_file;
      switch_to_file ((mainw->current_file=0),current_file);
      d_print (g_strdup_printf (_ ("rendered %d frames to new clip.\n"),cfile->frames));
      mainw->pre_src_file=mainw->current_file; // if a generator started playback, we will switch back to this file after
    }
    close_scrap_file();
  }
  else {
    retval=FALSE; // cancelled, so show the dialog again
    if (new_clip&&mainw->multitrack==NULL) {
      close_current_file();
      mainw->current_file=current_file;
    }
  }

  mainw->effects_paused=FALSE;
  deinit_render_effects();
  return retval;
}






void deal_with_render_choice (gboolean add_deinit) {
  // this is called from saveplay.c after record/playback ends
  // here we deal with the user's wishes as to how to deal with the recorded events

  // mainw->osc_block should be TRUE during all of this, so we don't have to contend with
  // any incoming network messages

  GtkWidget *e_rec_dialog;
  GtkWidget *elist_dialog;
  event_list_t *old_event_list;

  // record end
  mainw->record=FALSE;
  g_signal_handler_block(mainw->record_perf,mainw->record_perf_func);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mainw->record_perf),FALSE);
  g_signal_handler_unblock(mainw->record_perf,mainw->record_perf_func);
  
  if (add_deinit) add_effect_deinit_events (); // add deinit events for any effects that are left on

  if (cfile->event_list!=NULL) {
    event_list_free (cfile->event_list);
    cfile->event_list=NULL;
  }

  // TODO !!
  if (add_deinit) event_list_add_offset (U_SEC/cfile->fps*(gdouble)(mainw->play_start-1.),mainw->event_list);

  if (prefs->rec_opts&REC_EFFECTS||(prefs->rec_opts&REC_CLIPS&&mainw->clip_switched)) {
    do {
      e_rec_dialog=events_rec_dialog();
      gtk_widget_show (e_rec_dialog);
      gtk_dialog_run (GTK_DIALOG (e_rec_dialog));
      gtk_widget_destroy (e_rec_dialog);
      while (g_main_context_iteration (NULL,FALSE));
      switch (render_choice) {
      case RENDER_CHOICE_DISCARD:
	event_list_free (mainw->event_list);
	mainw->event_list=NULL;
	cfile->redoable=FALSE;
	close_scrap_file();
	sensitize();
	break;
      case RENDER_CHOICE_PREVIEW:
	// preview
	cfile->event_list=mainw->event_list; // TODO !!
	cfile->next_event=cfile->event_list->first;
	mainw->is_rendering=TRUE;
	on_preview_clicked (NULL,NULL);
	deinit_render_effects();
	mainw->is_processing=mainw->is_rendering=FALSE;
	cfile->event_list=NULL;
	cfile->next_event=NULL;
	break;
      case RENDER_CHOICE_NEW_CLIP:
	if (!render_to_clip (TRUE)) render_choice=RENDER_CHOICE_PREVIEW;
	mainw->is_rendering=FALSE;
	break;
      case RENDER_CHOICE_SAME_CLIP:
	if (!render_to_clip (FALSE)) render_choice=RENDER_CHOICE_PREVIEW;
	mainw->is_rendering=FALSE;
	break;
      case RENDER_CHOICE_MULTITRACK:
	if (on_multitrack_activate (NULL, (gpointer)mainw->event_list)) {
	  mainw->event_list=NULL;
	}
	else render_choice=RENDER_CHOICE_PREVIEW;
	break;
      case RENDER_CHOICE_EVENT_LIST:
	elist_dialog=create_event_list_dialog(mainw->event_list);
	gtk_dialog_run(GTK_DIALOG(elist_dialog));
	gtk_widget_destroy(elist_dialog);
	render_choice=RENDER_CHOICE_PREVIEW;
	break;
      }
      if (mainw->files[mainw->current_file]!=NULL) {
	cfile->event_list=NULL;
	cfile->next_event=NULL;
      }
    } while (render_choice==RENDER_CHOICE_PREVIEW);
  }
  else {
    // now we should have only frame events for a single clip
    if (mainw->event_list!=NULL) {
      if (!(prefs->rec_opts&REC_FPS)||(prefs->rec_opts&REC_CLIPS&&mainw->clip_switched)) {
	old_event_list=mainw->event_list;
	event_list_replace_events(mainw->event_list,quantise_events(mainw->event_list,cfile->fps,TRUE));
	cfile->variable_fps=FALSE;
	if (mainw->event_list==NULL) {
	  do_memory_error_dialog();
	  // we can fail to allocate memory in quantise events,
	  // and sensitize will already have made this TRUE
	  gtk_widget_set_sensitive(mainw->reorder,FALSE);
	}
	event_list_free (old_event_list);
      }
      else cfile->variable_fps=TRUE;
      if (cfile->undo_action==UNDO_RESAMPLE||cfile->undo_action==UNDO_REORDER) {
	set_undoable (NULL,FALSE);
      }
      if (mainw->event_list!=NULL) {
	// for now we just hand event_list off to cfile - in future we will do more interesting things..
	if (cfile->event_list!=NULL) event_list_free (cfile->event_list);
	cfile->event_list=mainw->event_list;
	mainw->event_list=NULL;
	gtk_widget_set_sensitive (mainw->reorder, TRUE);
	gtk_widget_set_sensitive (mainw->resample_video, FALSE);
	d_print(_ ("done.\n"));
      }
      else d_print (_ ("failed.\n"));
    }
    else {
      d_print (_ ("nothing recorded.\n"));
      if (cfile->event_list!=NULL) {
	event_list_free (cfile->event_list);
	cfile->event_list=NULL;
      }
      cfile->next_event=NULL;
    }
  }
}



void 
on_event_dialog_ok_clicked (GtkButton *button, gpointer user_data) {
  gtk_dialog_response (GTK_DIALOG (gtk_widget_get_toplevel(GTK_WIDGET(button))), GTK_RESPONSE_OK);
}


enum {
  TITLE_COLUMN,
  KEY_COLUMN,
  VALUE_COLUMN,
  NUM_COLUMNS
};



GtkWidget *create_event_list_dialog (event_list_t *event_list) {
 GtkTreeStore *gtkstore;
 GtkTreeIter iter1,iter2,iter3;

 gchar *string;
 int intval;
 void *voidval;
 event_t *event;

 GtkWidget *event_dialog;
 GtkWidget *tree;
 GtkWidget *table;
 GtkWidget *top_vbox;
 GtkWidget *label;
 GtkWidget *ok_button;
 GtkWidget *hbuttonbox;
 GtkWidget *scrolledwindow;
 GtkCellRenderer *renderer;
 GtkTreeViewColumn *column;

 gint winsize_h=mainw->scr_width-100;
 gint winsize_v=mainw->scr_height-100;

 int i,j,num_elems,seed_type,hint,error;
 weed_timecode_t tc,tc_secs;
 gchar *strval;

 char **propnames;

 gint rows,currow=0;
 rows=count_events(event_list,TRUE);

 event=event_list->first;
 
 event_dialog = gtk_dialog_new ();
 gtk_widget_modify_bg(event_dialog, GTK_STATE_NORMAL, &palette->menu_and_bars);
 gtk_window_set_title (GTK_WINDOW (event_dialog), _("LiVES: Event list"));
 gtk_window_add_accel_group (GTK_WINDOW (event_dialog), mainw->accel_group);
 
 top_vbox=GTK_DIALOG(event_dialog)->vbox;

 table = gtk_table_new (rows, 6, TRUE);
 gtk_widget_show (table);

 while (event!=NULL) {
   gtkstore = gtk_tree_store_new (NUM_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

   gtk_tree_store_append (gtkstore, &iter1, NULL);  /* Acquire an iterator */
   gtk_tree_store_set (gtkstore, &iter1, TITLE_COLUMN, "Properties", -1);
   
   propnames=weed_plant_list_leaves (event->event);
   
   for (i=0;propnames[i]!=NULL;i++) {
     if (!strcmp(propnames[i],"type")||!strcmp(propnames[i],"hint")||!strcmp(propnames[i],"timecode")) continue;
     gtk_tree_store_append (gtkstore, &iter2, &iter1);  /* Acquire a child iterator */
   
     num_elems=weed_leaf_num_elements (event->event,propnames[i]);
     seed_type=weed_leaf_seed_type (event->event,propnames[i]);

     for (j=0;j<num_elems;j++) {
       if (j>0) break; // for now, until the weed fns are better defined - TODO
       switch (seed_type) {
       case WEED_SEED_INT:
       case WEED_SEED_BOOLEAN:
	 intval=weed_get_int_value (event->event,propnames[i],&error);
	 strval=g_strdup_printf("%d",intval);
	 break;
       case WEED_SEED_STRING:
	 string=weed_get_string_value (event->event,propnames[i],&error);
	 strval=g_strdup(string);
	 weed_free(string);
	 break;
       case WEED_SEED_VOIDPTR:
	 voidval=weed_get_voidptr_value (event->event,propnames[i],&error);
	 strval=g_strdup_printf("%p",voidval);
	 break;
       default:
	 strval=g_strdup("???");
	 break;
       }

       if (j==0) gtk_tree_store_set (gtkstore, &iter2, KEY_COLUMN, propnames[i], VALUE_COLUMN, strval, -1);
       else {
	 gtk_tree_store_append (gtkstore, &iter3, &iter2);  /* Acquire a child iterator */
	 gtk_tree_store_set (gtkstore, &iter3, VALUE_COLUMN, strval, -1);
       }
       g_free (strval);
     }
     free(propnames[i]);
   }

   free (propnames);

   // now add the new treeview
   
   // timecode
   tc=get_event_timecode (event->event);
   tc_secs=tc/U_SECL;
   tc-=tc_secs*U_SECL;
   label = gtk_label_new (g_strdup_printf(("Timecode=%lld.%lld"),tc_secs,tc));
   gtk_widget_show (label);

   gtk_table_attach (GTK_TABLE (table), label, 0, 1, currow, currow+1,
		     (GtkAttachOptions) (GTK_EXPAND),
		     (GtkAttachOptions) (0), 0, 0);

   // event type
   hint=get_event_hint (event->event);
   switch (hint) {
   case LIVES_EVENT_HINT_FRAME:
     label = gtk_label_new ("Frame");
     break;
   case LIVES_EVENT_HINT_EFFECT_INIT:
     label = gtk_label_new ("Effect on");
     break;
   case LIVES_EVENT_HINT_EFFECT_DEINIT:
     label = gtk_label_new ("Effect off");
     break;
   case LIVES_EVENT_HINT_PARAM_CHANGE:
     label = gtk_label_new ("Parameter change");
     break;
   case LIVES_EVENT_HINT_EFFECT_MAP:
     label = gtk_label_new ("Effect map");
     break;
   }
   
   gtk_widget_show (label);
   gtk_table_attach (GTK_TABLE (table), label, 1, 2, currow, currow+1,
		     (GtkAttachOptions) (GTK_EXPAND),
		     (GtkAttachOptions) (0), 0, 0);

   
   // event id
   label = gtk_label_new (g_strdup_printf(("Event id=%p"),event));
   gtk_widget_show (label);
   gtk_table_attach (GTK_TABLE (table), label, 2, 3, currow, currow+1,
		     (GtkAttachOptions) (GTK_EXPAND),
		     (GtkAttachOptions) (0), 0, 0);
   
   // properties
   tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (gtkstore));
   gtk_widget_modify_base(tree, GTK_STATE_NORMAL, &palette->menu_and_bars);
   
   renderer = gtk_cell_renderer_text_new ();
   column = gtk_tree_view_column_new_with_attributes (NULL,
						      renderer,
						      "text", TITLE_COLUMN,
						      NULL);
   gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);
   
   renderer = gtk_cell_renderer_text_new ();
   column = gtk_tree_view_column_new_with_attributes ("Keys",
						      renderer,
						      "text", KEY_COLUMN,
						      NULL);
   gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);
   
   
   renderer = gtk_cell_renderer_text_new ();
   column = gtk_tree_view_column_new_with_attributes ("Values",
						      renderer,
						      "text", VALUE_COLUMN,
						      NULL);
   gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);
   gtk_widget_show (tree);

   gtk_table_attach (GTK_TABLE (table), tree, 3, 6, currow, currow+1,
		     (GtkAttachOptions) (GTK_FILL|GTK_EXPAND),
		     (GtkAttachOptions) (0), 0, 0);

   event=event->next;
   currow++;
 }

 scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
 gtk_widget_show (scrolledwindow);
 gtk_box_pack_start (GTK_BOX (top_vbox), scrolledwindow, TRUE, TRUE, 0);

 gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
 gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolledwindow), table);
 gtk_widget_set_size_request (scrolledwindow, winsize_h, winsize_v);
 
 if (palette->style&STYLE_1) {
   gtk_widget_modify_bg(gtk_bin_get_child (GTK_BIN (scrolledwindow)), GTK_STATE_NORMAL, &palette->normal_back);
 }
 
 gtk_viewport_set_shadow_type (GTK_VIEWPORT (gtk_bin_get_child (GTK_BIN (scrolledwindow))),GTK_SHADOW_IN);

 hbuttonbox = gtk_hbutton_box_new ();
 gtk_widget_show (hbuttonbox);

 gtk_box_pack_start (GTK_BOX (top_vbox), hbuttonbox, TRUE, TRUE, 0);

 gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbuttonbox), DEF_BUTTON_WIDTH, -1);
 gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox), GTK_BUTTONBOX_SPREAD);

 ok_button = gtk_button_new_with_mnemonic (_("Close _window"));
 gtk_widget_show (ok_button);
 gtk_container_add (GTK_CONTAINER (hbuttonbox), ok_button);

 GTK_WIDGET_SET_FLAGS (ok_button, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
 gtk_widget_grab_default (ok_button);
 
 g_signal_connect (GTK_OBJECT (ok_button), "clicked",
		   G_CALLBACK (on_event_dialog_ok_clicked),
		   NULL);

 gtk_window_maximize (GTK_WINDOW(event_dialog));

 return event_dialog;
}




void rdetw_spinh_changed (GtkSpinButton *spinbutton, gpointer user_data) {
  render_details *rdet=(render_details *)user_data;
  rdet->height=gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton));
}


void rdetw_spinw_changed (GtkSpinButton *spinbutton, gpointer user_data) {
  render_details *rdet=(render_details *)user_data;
  rdet->width=gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton));
}


void rdetw_spinf_changed (GtkSpinButton *spinbutton, gpointer user_data) {
  render_details *rdet=(render_details *)user_data;
  rdet->fps=gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinbutton));
}



GtkWidget *create_render_details_dialog (render_details *rdet) {
  // TODO - allow user to pick target encoder here, then LiVES can suggest a corrected frame size/fps

  GtkWidget *rdet_dialog;
  GtkWidget *label;
  GtkWidget *top_vbox;
  GtkWidget *hbox;
  GtkWidget *spinbutton;
  GtkObject *spinbutton_adj;
  GtkWidget *okbutton;
  GtkWidget *cancelbutton;
  GtkAccelGroup *rdet_accel_group;

  rdet->height=DEFAULT_FRAME_VSIZE;
  rdet->width=DEFAULT_FRAME_HSIZE;
  rdet->fps=prefs->default_fps;

  rdet_dialog = gtk_dialog_new ();
  gtk_widget_modify_bg(rdet_dialog, GTK_STATE_NORMAL, &palette->menu_and_bars);
  gtk_window_set_title (GTK_WINDOW (rdet_dialog), _("LiVES: New clip details"));
  gtk_container_set_border_width (GTK_CONTAINER (rdet_dialog), 10);
  gtk_window_set_default_size (GTK_WINDOW (rdet_dialog), 300, 200);
  gtk_window_set_transient_for(GTK_WINDOW(rdet_dialog),GTK_WINDOW(mainw->LiVES));
  gtk_window_set_modal (GTK_WINDOW (rdet_dialog), TRUE);
  rdet_accel_group = GTK_ACCEL_GROUP(gtk_accel_group_new ());
  gtk_window_add_accel_group (GTK_WINDOW (rdet_dialog), rdet_accel_group);

  top_vbox = GTK_DIALOG(rdet_dialog)->vbox;

  hbox = gtk_hbox_new (FALSE, 50);

  gtk_box_pack_start (GTK_BOX (top_vbox), hbox, TRUE, TRUE, 0);

  label = gtk_label_new_with_mnemonic (_ ("_Width"));

  spinbutton_adj = gtk_adjustment_new (rdet->width, 1., 100000., 1, 10, 10);

  spinbutton = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_adj), 1, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spinbutton),TRUE);
  gtk_entry_set_activates_default (GTK_ENTRY (spinbutton), TRUE);

  g_signal_connect_after (GTK_OBJECT (spinbutton), "value_changed",
			  G_CALLBACK (rdetw_spinw_changed),
			  rdet);

  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 10);
  gtk_box_pack_start (GTK_BOX (hbox), spinbutton, FALSE, FALSE, 10);
  
  label = gtk_label_new_with_mnemonic (_ ("_Height"));

  spinbutton_adj = gtk_adjustment_new (rdet->height, 1., 10000., 1, 10, 10);

  spinbutton = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_adj), 1, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spinbutton),TRUE);
  gtk_entry_set_activates_default (GTK_ENTRY (spinbutton), TRUE);

  g_signal_connect_after (GTK_OBJECT (spinbutton), "value_changed",
			  G_CALLBACK (rdetw_spinh_changed),
			  rdet);


  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 10);
  gtk_box_pack_start (GTK_BOX (hbox), spinbutton, FALSE, FALSE, 10);

  hbox = gtk_hbox_new (FALSE, 0);

  gtk_box_pack_start (GTK_BOX (top_vbox), hbox, TRUE, TRUE, 0);

  label = gtk_label_new_with_mnemonic (_ ("_Frames per second"));

  spinbutton_adj = gtk_adjustment_new (rdet->fps, .0001, FPS_MAX, 1, 10, 10);

  spinbutton = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_adj), 1, 3);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spinbutton),TRUE);
  gtk_entry_set_activates_default (GTK_ENTRY (spinbutton), TRUE);

  g_signal_connect_after (GTK_OBJECT (spinbutton), "value_changed",
			  G_CALLBACK (rdetw_spinf_changed),
			  rdet);

  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 10);
  gtk_box_pack_start (GTK_BOX (hbox), spinbutton, FALSE, FALSE, 10);

  cancelbutton = gtk_button_new_from_stock ("gtk-cancel");
  gtk_dialog_add_action_widget (GTK_DIALOG (rdet_dialog), cancelbutton, GTK_RESPONSE_CANCEL);
  GTK_WIDGET_SET_FLAGS (cancelbutton, GTK_CAN_FOCUS);

  okbutton = gtk_button_new_from_stock ("gtk-ok");
  gtk_dialog_add_action_widget (GTK_DIALOG (rdet_dialog), okbutton, GTK_RESPONSE_OK);
  GTK_WIDGET_SET_FLAGS (okbutton, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
  gtk_widget_grab_default (okbutton);

  gtk_widget_add_accelerator (cancelbutton, "activate", rdet_accel_group,
			      GDK_Escape, 0, 0);

  gtk_widget_show_all (rdet_dialog);
  return rdet_dialog;
}






