/* WEED is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   Weed is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this source code; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   Weed is developed by:

   Gabriel "Salsaman" Finch - http://lives.sourceforge.net


   mainly based on LiViDO, which is developed by:


   Niels Elburg - http://veejay.sf.net

   Gabriel "Salsaman" Finch - http://lives.sourceforge.net

   Denis "Jaromil" Rojo - http://freej.dyne.org

   Tom Schouten - http://zwizwa.fartit.com

   Andraz Tori - http://cvs.cinelerra.org

   reviewed with suggestions and contributions from:

   Silvano "Kysucix" Galliani - http://freej.dyne.org

   Kentaro Fukuchi - http://megaui.net/fukuchi

   Jun Iio - http://www.malib.net

   Carlo Prelz - http://www2.fluido.as:8080/

*/

/* (C) Gabriel "Salsaman" Finch, 2005 */


/* Get size_t and NULL from <stddef.h>.  */

#define __need_size_t
#define __need_NULL
#include <stddef.h>

#include "weed.h"

static inline size_t weed_strlen (const char *string) {
  size_t len=0;
  while (*(string++)!=0) len++;
  return len;
}

static inline char *weed_strdup (const char *string) {
  size_t len;
  char *ret=weed_malloc((len=weed_strlen(string))+1);
  weed_memcpy(ret,string,len);
  weed_memset(ret+len,0,1);
  return ret;
}

static inline int weed_strcmp (const char *st1, const char *st2) {
  while (!(*st1==0&&*st2==0)) {
    if (*(st1)==0||*(st2)==0||*(st1++)!=*(st2++)) return 1;
  }
  return 0;
}


static inline int weed_seed_is_ptr (int seed) {
 return (seed!=WEED_SEED_BOOLEAN&&seed!=WEED_SEED_INT&&seed!=WEED_SEED_DOUBLE&&seed!=WEED_SEED_STRING&&seed!=WEED_SEED_INT64)?1:0;
}

static inline size_t weed_seed_get_size (int seed, void *value) {
  return weed_seed_is_ptr (seed)?(sizeof(void *)):\
  (seed==WEED_SEED_BOOLEAN||seed==WEED_SEED_INT)?sizeof(int):\
  (seed==WEED_SEED_DOUBLE)?sizeof(double):\
  (seed==WEED_SEED_INT64)?sizeof(long long):\
  (seed==WEED_SEED_STRING)?weed_strlen((const char *)value):0;
}

static inline void weed_data_free(weed_data_t **data, int num_elems, int seed_type) {
  register int i;
  for (i=0;i<num_elems;i++) {
    if (!weed_seed_is_ptr(seed_type)||seed_type==WEED_SEED_STRING) weed_free(data[i]->value);
    weed_free(data[i]);
  }
  weed_free(data);
}

static inline weed_data_t **weed_data_new(int seed_type, int num_elems, void **value) {
  register int i;
  weed_data_t **data=NULL;
  size_t size;
  if (num_elems==0) return data;
  if ((data=(weed_data_t **)weed_malloc(num_elems*sizeof(weed_data_t *)))==NULL) return ((void *)0);
  for (i=0;i<num_elems;i++) {
    if ((data[i]=(weed_data_t *)weed_malloc(sizeof(weed_data_t)))==NULL) {
      weed_data_free(data,--i,seed_type);
      return NULL;
    }
    if (weed_seed_is_ptr(seed_type)) weed_memcpy(&data[i]->value,&value[i],sizeof(void *));
    else {
      if (seed_type==WEED_SEED_STRING) data[i]->value=weed_strdup(value[i]);
      else if ((data[i]->value=weed_malloc((size=weed_seed_get_size(seed_type,value[i]))))!=NULL) weed_memcpy(data[i]->value,&value[i],size);
      if (data[i]->value==NULL) { // memory error
	weed_data_free(data,--i,seed_type);
	return NULL;
      }
    }
    data[i]->size=weed_seed_get_size(seed_type,data[i]->value);
  }
  return data;
}

static inline weed_leaf_t *weed_find_leaf(weed_plant_t *leaf, const char *key) {
 while (leaf!=NULL) {
  if (!weed_strcmp((char *)leaf->key,(char *)key)) return leaf;
  leaf=leaf->next;
 }
 return NULL;
}

static inline void weed_leaf_free(weed_leaf_t *leaf) {
  weed_data_free (leaf->data,leaf->num_elements,leaf->seed_type);
  weed_free ((char *)leaf->key);
  weed_free (leaf);
}

static inline weed_leaf_t *weed_leaf_new(const char *key, int seed) {
  weed_leaf_t *leaf=(weed_leaf_t *)weed_malloc(sizeof(weed_leaf_t));
  if (leaf==NULL) return NULL;
  if ((leaf->key=weed_strdup(key))==NULL) {
    weed_free (leaf);
    return NULL;
  }
  leaf->seed_type=seed;
  leaf->num_elements=0;
  leaf->data=NULL;
  leaf->next=NULL;
  leaf->flags=0;
  return leaf;
}

static inline void weed_leaf_append(weed_plant_t *leaf, weed_leaf_t *newleaf) {
  weed_leaf_t *leafnext;
  while (leaf!=NULL) {
    leafnext=leaf->next;
    if (leafnext==NULL) {
      leaf->next=newleaf;
      return;
    }
    leaf=leafnext;
  }
}

void weed_plant_free(weed_plant_t *leaf) {
  weed_leaf_t *leafnext;
  while (leaf!=NULL) {
    leafnext=leaf->next;
    weed_leaf_free (leaf);
    leaf=leafnext;
  }
}

int weed_leaf_set_flags(weed_plant_t *plant, const char *key, int flags) {
  weed_leaf_t *leaf=weed_find_leaf (plant, key);
  if (leaf==NULL) return WEED_ERROR_NOSUCH_LEAF;
  leaf->flags=flags;
  return WEED_NO_ERROR;
}

weed_plant_t *weed_plant_new(int plant_type) {
  weed_leaf_t *leaf;
  void *vdplant_type=(void *)plant_type;
  if ((leaf=weed_leaf_new("type",WEED_SEED_INT))==NULL) return NULL;
  if ((leaf->data=(weed_data_t **)weed_data_new (WEED_SEED_INT,1,(void **)&vdplant_type))==NULL) {
    weed_free ((char *)leaf->key);
    weed_free (leaf);
    return NULL;
  }
  leaf->next=NULL;
  weed_leaf_set_flags (leaf,"type",WEED_LEAF_READONLY_PLUGIN|WEED_LEAF_READONLY_HOST);
  return leaf;
}

char **weed_plant_list_leaves(weed_plant_t *plant) {
  weed_leaf_t *leaf=plant;
  char **leaflist;
  register int i=1;
  for (;leaf!=NULL;i++) {
    leaf=leaf->next;
  }
  if ((leaflist=(char **)weed_malloc(i*sizeof(char *)))==NULL) return NULL;
  i=0;
  for (leaf=plant;leaf!=NULL;leaf=leaf->next) {
    if ((leaflist[i]=weed_strdup(leaf->key))==NULL) {
      for (--i;i>=0;i--) weed_free (leaflist[i]);
      weed_free (leaflist);
      return NULL;
    }
    i++;
  }
  leaflist[i]=NULL;
  return leaflist;
}

int weed_leaf_set_caller(weed_plant_t *plant, const char *key, int seed_type, int num_elems, void *value, int caller) {
  weed_data_t **data=NULL;
  weed_leaf_t *leaf=weed_find_leaf (plant,key);
  if (leaf==NULL) {
    if ((leaf=weed_leaf_new (key,seed_type))==NULL) return WEED_ERROR_MEMORY_ALLOCATION;
    weed_leaf_append (plant,leaf);
  }
  else {
    if ((caller==WEED_CALLER_PLUGIN&&leaf->flags&WEED_LEAF_READONLY_PLUGIN)||(caller==WEED_CALLER_HOST&&leaf->flags&WEED_LEAF_READONLY_HOST)) return WEED_ERROR_LEAF_READONLY;
    if (seed_type!=leaf->seed_type) return WEED_ERROR_WRONG_SEED_TYPE;
    weed_data_free (leaf->data,leaf->num_elements,seed_type);
    leaf->data=NULL;
  }
  leaf->num_elements=0;
  if (num_elems>0&&(data=weed_data_new (seed_type,num_elems,(void **)value))==NULL) {
    return WEED_ERROR_MEMORY_ALLOCATION;
  }
  leaf->data=data;
  leaf->num_elements=num_elems;
  return WEED_NO_ERROR;
}

int weed_leaf_get(weed_plant_t *plant, const char *key, int idx, void *value) {
  size_t size;
  weed_leaf_t *leaf=weed_find_leaf (plant,key);
  if (leaf==NULL||idx>leaf->num_elements) return WEED_ERROR_NOSUCH_LEAF;
  if (value==NULL) return WEED_NO_ERROR;
  if (weed_seed_is_ptr (leaf->seed_type)) weed_memcpy (value,&leaf->data[idx]->value,sizeof(void *));
  else {
    if (leaf->seed_type==WEED_SEED_STRING) {
      char **valuecharptrptr=(char **)value;
      weed_memcpy(*valuecharptrptr,leaf->data[idx]->value,(size=weed_strlen(leaf->data[idx]->value)));
      weed_memset(*valuecharptrptr+size,0,1);
    }
    else weed_memcpy (value,leaf->data[idx]->value,weed_seed_get_size(leaf->seed_type,leaf->data[idx]->value));
  }
  return WEED_NO_ERROR;
}

int weed_leaf_num_elements(weed_plant_t *plant, const char *key) {
  weed_leaf_t *leaf=weed_find_leaf (plant, key);
  if (leaf==NULL) return 0;
  return leaf->num_elements;
}

size_t weed_leaf_element_size(weed_plant_t *plant, const char *key, int idx) {
  weed_leaf_t *leaf=weed_find_leaf (plant, key);
  if (leaf==NULL||idx>leaf->num_elements) return 0;
  return weed_seed_get_size(leaf->seed_type,leaf->data[idx]->value);
}

int weed_leaf_seed_type(weed_plant_t *plant, const char *key) {
  weed_leaf_t *leaf=weed_find_leaf (plant, key);
  if (leaf==NULL) return 0;
  return leaf->seed_type;
}

int weed_leaf_get_flags(weed_plant_t *plant, const char *key) {
  weed_leaf_t *leaf=weed_find_leaf (plant, key);
  if (leaf==NULL) return 0;
  return leaf->flags;
}

